from __future__ import annotations

import ast
import sys
from types import ModuleType
from typing import Callable

from _frozen_importlib_external import PathFinder  # pyright: reportMissingImports=false

import deal

from ._state import state


def get_name(expr) -> str | None:
    if isinstance(expr, ast.Name):
        return expr.id
    if isinstance(expr, ast.Attribute):
        left = get_name(expr.value)
        if left is None:
            return None
        return left + '.' + expr.attr
    return None


class DealFinder(PathFinder):
    @classmethod
    def find_spec(cls, *args, **kwargs):
        spec = super().find_spec(*args, **kwargs)
        if spec is not None:  # pragma: no cover
            spec.loader = DealLoader(spec.loader)
        return spec


class DealLoader:
    __slots__ = ('_loader', )

    def __init__(self, loader) -> None:
        self._loader = loader

    def __getattr__(self, name: str):
        return getattr(self._loader, name)

    def exec_module(self, module: ModuleType) -> None:
        if not hasattr(self._loader, 'get_source'):
            return self._loader.exec_module(module)

        # get nodes with module-level contracts from the source code
        source = self._loader.get_source(module.__name__)
        if source is None:
            return self._loader.exec_module(module)
        tree = ast.parse(source)
        nodes = self._get_contracts(tree=tree)
        if not nodes:
            return self._loader.exec_module(module)

        # convert contracts nodes into real contracts
        contracts = []
        for node in nodes:
            contract = self._exec_contract(node=node)
            if contract is None:
                msg = f'unsupported contract: {ast.dump(node)}'
                raise RuntimeError(msg)
            contracts.append(contract)

        # execute module with contracts
        wrapped = deal.chain(*contracts)(self._loader.exec_module)
        wrapped(module)

    @staticmethod
    def _get_contracts(tree: ast.Module) -> list[ast.expr]:
        for node in tree.body:
            if type(node) is not ast.Expr:
                continue
            if type(node.value) is not ast.Call:
                continue
            if get_name(node.value.func) != 'deal.module_load':
                continue
            return node.value.args
        return []

    @classmethod
    def _exec_contract(cls, node: ast.AST) -> Callable | None:
        """Get AST node and return a contract function
        """
        if isinstance(node, ast.Call) and not node.keywords:
            try:
                args = [ast.literal_eval(arg) for arg in node.args]
            except ValueError:
                return None
            func = cls._exec_contract(node.func)
            if not func:
                return None
            return func(*args)

        if not isinstance(node, ast.Attribute):
            return None
        if node.value.id != 'deal':
            return None
        contract = getattr(deal, node.attr, None)
        if contract is None:
            return None
        return contract


def module_load(*contracts) -> None:
    """
    Specify contracts that will be checked at module import time.
    Keep in mind that [deal.activate](#deal.activate) must be called
    before importing a module with `module_load` contract.

    ```pycon
    >>> import deal
    >>> deal.module_load(deal.has(), deal.safe)

    ```

    See [Contracts for importing modules][module_load]
    documentation for more details.

    [module_load]: https://deal.readthedocs.io/details/module_load.html
    """
    if not state.debug:
        return
    if not contracts:
        raise RuntimeError('no contracts specified')
    if DealFinder not in sys.meta_path:
        msg = 'deal.activate must be called '
        msg += 'before importing anything with deal.module_load contract'
        raise RuntimeError(msg)


def activate() -> bool:
    """Activate module-level checks.

    This function must be called before importing anything
    with [deal.module_load](#deal.module_load) contract.
    Otherwise, the contract won't be executed.

    The best practice is to place it in `__init__.py` of your project:

    ```pycon
    >>> import deal
    >>> deal.activate()

    ```

    See [Contracts for importing modules][module_load]
    documentation for more details.

    [module_load]: https://deal.readthedocs.io/details/module_load.html
    """
    if not state.debug:
        return False
    if DealFinder in sys.meta_path:
        return False
    index = sys.meta_path.index(PathFinder)
    sys.meta_path[index] = DealFinder
    return True


def deactivate() -> bool:
    """used in tests
    """
    if DealFinder not in sys.meta_path:
        return False
    index = sys.meta_path.index(DealFinder)
    sys.meta_path[index] = PathFinder
    return True



from __future__ import annotations

import ast
import builtins
import re
from inspect import cleandoc
from typing import Iterator

from .._contract import Category
from .._stub import StubsManager
from .common import TOKENS, Extractor, Token, get_full_name, get_name, get_stub, infer
from .contracts import get_contracts


try:
    import astroid
except ImportError:
    astroid = None
try:
    import docstring_parser
except ImportError:
    docstring_parser = None


get_exceptions = Extractor()
REX_GOOGLE_SECTION = re.compile(r'[A-Z][a-z]+:\s*')


@get_exceptions.register(*TOKENS.ASSERT)
def handle_assert(expr, **kwargs) -> Token | None:
    return Token(value=AssertionError)


# explicit raise
@get_exceptions.register(*TOKENS.RAISE)
def handle_raise(expr: ast.Raise, **kwargs) -> Token | None:
    if expr.exc is None:
        return None
    name = get_name(expr.exc)
    if not name:
        # raised a value, too tricky
        if not isinstance(expr.exc, TOKENS.CALL):
            return None
        # raised an instance of an exception
        name = get_name(expr.exc.func)
        if not name or name[0].islower():
            return None
    exc = getattr(builtins, name, name)
    return Token(value=exc, line=expr.exc.lineno, col=expr.exc.col_offset)


# division by zero
@get_exceptions.register(*TOKENS.BIN_OP)
def handle_bin_op(expr: ast.BinOp | astroid.BinOp, **kwargs) -> Token | None:
    if isinstance(expr.op, ast.Div) or expr.op == '/':
        if astroid is not None and isinstance(expr.right, astroid.NodeNG):
            guesses = infer(expr=expr.right)
            for guess in guesses:
                if type(guess) is not astroid.Const:
                    continue
                return Token(value=ZeroDivisionError, col=expr.right.col_offset)
        if isinstance(expr.right, ast.Constant) and str(expr.right.value) == '0':
            return Token(value=ZeroDivisionError, col=expr.right.col_offset)
    return None


@get_exceptions.register(*TOKENS.CALL)
def handle_call(
    expr: ast.Call | astroid.Call,
    dive: bool = True,
    stubs: StubsManager | None = None,
) -> Iterator[Token]:
    # exit()
    name = get_name(expr.func)
    if name and name == 'exit':
        yield Token(value=SystemExit)
    # sys.exit()
    if isinstance(expr.func, TOKENS.ATTR):
        name = get_name(expr.func)
        if name and name == 'sys.exit':
            yield Token(value=SystemExit)

    stubs_found = False
    if astroid is not None and type(expr) is astroid.Call and stubs is not None:
        for token in _exceptions_from_stubs(expr=expr, stubs=stubs):
            stubs_found = True
            yield token

    # Infer function call and check the function body for raises.
    # Do not dive into called function if we already found stubs for it.
    if not stubs_found and dive:
        yield from _exceptions_from_func(expr=expr)


def _exceptions_from_stubs(expr: astroid.Call, stubs: StubsManager) -> Iterator[Token]:
    for value in infer(expr=expr.func):
        if type(value) is not astroid.FunctionDef:
            continue
        module_name, func_name = get_full_name(expr=value)
        stub = get_stub(module_name=module_name, expr=value, stubs=stubs)
        if stub is None:
            continue
        names = stub.get(func=func_name, contract=Category.RAISES)
        for name in names:
            name = getattr(builtins, name, name)
            yield Token(value=name)


def _exceptions_from_func(expr: ast.Call | astroid.Call) -> Iterator[Token]:
    if astroid is None:  # pragma: no-astroid
        return
    for value in infer(expr.func):
        if type(value) is not astroid.FunctionDef:
            continue

        # recursively infer exceptions from the function body
        for error in get_exceptions(body=value.body, dive=False):
            yield Token(value=error.value)

        # get explicitly specified exceptions from `@deal.raises`
        name: str | None
        for contract in get_contracts(value):
            if contract.name != 'raises':
                continue
            for arg in contract.args:
                name = get_name(arg)
                if name is None:
                    continue
                name = getattr(builtins, name, name)
                yield Token(value=name)

        # get exceptions from the docstring
        name: str
        if value.doc_node is not None:
            for name in _excs_from_doc(value.doc_node.value):
                name = getattr(builtins, name, name)
                yield Token(value=name)
    return None


# TODO: use it on the target function docstring too
def _excs_from_doc(doc: str) -> Iterator[str]:
    if docstring_parser is not None:
        parsed = docstring_parser.parse(doc)
        for exc_info in parsed.raises:
            if exc_info.type_name:
                yield exc_info.type_name
        return

    google_section = ''
    numpy_section = ''
    google_section_indent = 4
    numpy_section_indent = 0
    lines = cleandoc(doc).splitlines() + ['']
    for line, next_line in zip(lines, lines[1:]):
        words = line.split()
        if not words:
            continue
        indent = _get_indent(line)
        is_numpy_header = _is_header_highlight(next_line)

        # sphinx and epydoc docstring
        if len(words) >= 2 and words[0] in (':raises', '@raise'):
            yield words[1].rstrip(':')
            continue

        # google docstring
        if REX_GOOGLE_SECTION.fullmatch(line) and not is_numpy_header:
            google_section = line.strip().rstrip(':').lower()
            google_section_indent = _get_indent(next_line)
            continue
        if google_section == 'raises' and indent == google_section_indent:
            yield words[0].rstrip(':')
            continue

        # numpy docstring
        next_line = next_line.strip()
        if is_numpy_header:
            numpy_section = line.strip().rstrip(':').lower()
            continue
        if _is_header_highlight(line):
            numpy_section_indent = _get_indent(next_line)
            continue
        if numpy_section == 'raises' and indent == numpy_section_indent:
            yield line.rstrip()


def _get_indent(line: str) -> int:
    return len(line) - len(line.lstrip())


def _is_header_highlight(line: str) -> bool:
    line = line.strip()
    return len(set(line)) == 1 and line[0] in '-+'



import ast

import pytest

from deal.linter._extractors import get_asserts


try:
    import astroid
except ImportError:
    astroid = None


@pytest.mark.parametrize('text, expected', [
    ('assert 1', ()),
    ('assert True', ()),
    ('assert "abc"', ()),
    ('assert "abc", "do not panic"', ()),

    ('assert 0', (0, )),
    ('assert False', (False, )),
    ('assert ""', ('', )),
    ('assert []', ([], )),

    ('assert 3 - 2', ()),
    # ('assert 2 - 2', (0, )),
    ('assert a', ()),           # ignore uninferrable names
    ('a = object()\nassert a', ()),
])
def test_get_asserts_simple(text, expected):
    if astroid is not None:
        tree = astroid.parse(text)
        print(tree.repr_tree())
        returns = tuple(r.value for r in get_asserts(body=tree.body))
        assert returns == expected

    ast_tree = ast.parse(text)
    print(ast.dump(ast_tree))
    returns = tuple(r.value for r in get_asserts(body=ast_tree.body))
    assert returns == expected



from argparse import ArgumentParser

from deal._cli import main
from deal._cli._base import Command


class FakeCommand(Command):
    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument('args', nargs='*')

    def __call__(self, args) -> int:
        assert args.args == ['a1', 'a2']
        return 13


def test_calls_command():
    result = main(argv=['fake', 'a1', 'a2'], commands=dict(fake=FakeCommand))
    assert result == 13


def test_unknown_command(capsys):
    result = main(argv=['unknown', 'a1', 'a2'], commands=dict(fake=FakeCommand))
    assert result == 2
    captured = capsys.readouterr()
    exp1 = "invalid choice: 'unknown' (choose from fake)"
    exp2 = "invalid choice: 'unknown' (choose from 'fake')"
    assert exp1 in captured.err or exp2 in captured.err


def test_no_command_specified(capsys):
    result = main(argv=[], commands=dict(fake=FakeCommand))
    assert result == 2
    captured = capsys.readouterr()
    assert 'positional arguments:' in captured.out
    assert 'fake' in captured.out



#!/usr/bin/env python3
import os
import sys
from datetime import date
from pathlib import Path

import sphinx_rtd_theme
from m2r2 import MdInclude, convert
from sphinx.application import Sphinx

import deal


sys.path.append(os.path.abspath('../'))
extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.doctest',
    'sphinx.ext.todo',
    'sphinx.ext.coverage',
    'sphinx.ext.viewcode',
    'sphinx.ext.githubpages',
    'sphinx.ext.autosectionlabel',
    'myst_parser',
]

templates_path = ['_templates']
source_suffix = ['.rst', '.md']
master_doc = 'index'

project = 'Deal'
copyright = '{}, Gram (@orsinium)'.format(date.today().year)
author = 'Gram (@orsinium)'

version = ''
release = ''

language = 'en'
exclude_patterns: list = []
todo_include_todos = True

pygments_style = 'sphinx'
html_theme = 'sphinx_rtd_theme'
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]
html_static_path = [str(Path(__file__).parent.parent / 'assets')]
html_logo = str(Path(__file__).parent.parent / 'logo.png')
html_theme_options = {
    'logo_only': True,
    'display_version': False,
    'style_external_links': False,
    'style_nav_header_background': '#2c3e50',
}


# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'dealdoc'


# -- Options for LaTeX output ---------------------------------------------

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
    (master_doc, 'deal.tex', 'Deal Documentation',
     '@orsinium', 'manual'),
]


# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [(master_doc, 'deal', 'Deal Documentation', [author], 1)]


# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
    (master_doc, 'deal', 'Deal Documentation',
     author, 'Deal', 'One line description of project.', 'Miscellaneous'),
]

autosectionlabel_prefix_document = True
add_function_parentheses = False


def autodoc_process(app, what, name, obj, options, lines):
    if not lines:
        return None
    text = convert('\n'.join(lines))
    lines[:] = text.split('\n')


def autodoc_signature(
    app, what, name, obj, options,
    signature: str, return_annotation: str,
):
    sig = signature
    ret = return_annotation
    if sig:
        sig = sig.replace('deal._runtime._decorators.', '')
        sig = sig.replace('Union[Exception, Type[Exception]]', 'Exception')
    if ret:
        ret = ret.replace('deal._runtime._decorators.', '')
        ret = ret.replace('deal._runtime.dispatch.', '')
    return (sig, ret)


def setup(app: Sphinx):
    # from m2r2 source code to make `mdinclude` work
    app.add_config_value('no_underscore_emphasis', False, 'env')
    app.add_config_value('m2r_parse_relative_links', False, 'env')
    app.add_config_value('m2r_anonymous_references', False, 'env')
    app.add_config_value('m2r_disable_inline_math', False, 'env')
    app.add_config_value('m2r_use_mermaid', False, 'env')
    app.add_directive('mdinclude', MdInclude)

    app.connect('autodoc-process-docstring', autodoc_process)
    app.connect('autodoc-process-signature', autodoc_signature)

    deal.autodoc(app)



from __future__ import annotations

from typing import TYPE_CHECKING

from .common import TOKENS, Extractor, Token
from .value import UNKNOWN, get_value


if TYPE_CHECKING:
    import ast

    import astroid


get_asserts = Extractor()


@get_asserts.register(*TOKENS.ASSERT)
def handle_assert(expr: ast.Assert | astroid.Assert) -> Token | None:
    value = get_value(expr=expr.test, allow_inference=False)
    if value is UNKNOWN:
        return None
    if value:
        return None
    return Token(value=value, line=expr.lineno, col=expr.col_offset + 7)



from __future__ import annotations

import ast
from contextlib import suppress

from .common import infer


try:
    import astroid
except ImportError:
    astroid = None

UNKNOWN = object()


def get_value(expr: ast.AST | astroid.NodeNG, allow_inference: bool = True) -> object:
    if isinstance(expr, ast.AST):
        with suppress(ValueError, SyntaxError):
            return ast.literal_eval(expr)
        return UNKNOWN
    if astroid is None:  # pragma: no-astroid
        return UNKNOWN

    if isinstance(expr, astroid.NodeNG):
        # AttributeError: 'AsStringVisitor3' object has no attribute 'visit_unknown'
        with suppress(AttributeError):  # pragma: no cover
            renderred = expr.as_string()
            with suppress(ValueError, SyntaxError):
                return ast.literal_eval(renderred)

    value = _parse_collections(expr)
    if value is not UNKNOWN:
        return value

    if allow_inference:
        for parent_expr in infer(expr):
            if parent_expr == expr:  # avoid recursion
                continue
            value = get_value(parent_expr)
            if value is not UNKNOWN:
                return value
    return UNKNOWN


def _parse_collections(expr: astroid.NodeNG) -> object:
    if not isinstance(expr, (astroid.List, astroid.Set, astroid.Tuple)):
        return UNKNOWN

    result = []
    for element_expr in expr.elts:
        value = get_value(expr=element_expr)
        if value is UNKNOWN:
            return UNKNOWN
        result.append(value)

    if type(expr) is astroid.Tuple:
        return tuple(result)
    if type(expr) is astroid.Set:
        return set(result)
    return result



import doctest
import re
import sys
from itertools import chain
from unittest.mock import Mock

import pytest

import deal._runtime
import deal._testing
from deal._state import state


try:
    import astroid
except ImportError:
    astroid = None


rex = re.compile(r'deal\.(\w*Error)')


class Checker(doctest.OutputChecker):
    def __init__(self):
        self.diff = []

    def check_output(self, want: str, got: str, optionflags: int) -> bool:
        got = rex.sub(r'\1', got)
        ok = super().check_output(want=want, got=got, optionflags=optionflags)
        if not ok:
            self.diff.append((got, want))
        return ok


finder = doctest.DocTestFinder(exclude_empty=True)


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
@pytest.mark.parametrize('test', chain(
    finder.find(deal._runtime),
    finder.find(deal._testing),
))
def test_doctest(test):
    state.color = False
    sys.modules['atheris'] = Mock()
    try:
        runner = doctest.DocTestRunner(checker=Checker(), optionflags=doctest.ELLIPSIS)
        runner.run(test)
        result = runner.summarize(verbose=False)
        if result.failed:
            print('Kinda diff:')
            print(*runner._checker.diff, sep='\n')
        assert not result.failed
    finally:
        state.color = True
        del sys.modules['atheris']



import pytest

import deal
from deal._state import state


@deal.pre(lambda x: x == 3)
def double3(x: int) -> int:
    return 6


@deal.pre(lambda x: x == 4)
def double4(x: int) -> int:
    return 8


def double(x: int) -> int:
    """Documentation"""
    raise NotImplementedError


def test_docstring():
    f = deal.dispatch(double)
    f.register(double3)
    assert f.__doc__ == 'Documentation'


def test_match():
    f = deal.dispatch(double)
    f.register(double3)
    f.register(double4)
    assert f(3) == 6
    assert f(4) == 8


def test_no_match():
    f = deal.dispatch(double)
    f.register(double3)
    f.register(double4)
    with pytest.raises(deal.NoMatchError) as exc_info:
        f(2)
    state.color = False
    exp = 'expected x == 3 (where x=2); expected x == 4 (where x=2)'
    assert str(exc_info.value) == exp
    state.color = True


def test_match_default():
    f = deal.dispatch(double)
    f.register(double3)
    f.register(double4)
    f.register(lambda x: x * 2)
    assert f(10) == 20


def test_propagate_pre_contract_error():
    f = deal.dispatch(double)
    f.register(double3)

    @deal.pre(lambda: False)
    def bad_func() -> int:
        return 0

    @f.register
    @deal.pre(lambda x: x == 4)
    def _double4(x: int) -> int:
        return bad_func()

    assert f(3) == 6
    with pytest.raises(deal.PreContractError) as exc_info:
        f(4)
    assert exc_info.value.source == 'False'


def test_propagate_pre_contract_error_from_default():
    f = deal.dispatch(double)
    f.register(double3)

    @deal.pre(lambda: False)
    def bad_func() -> int:
        return 0

    @f.register
    def _double(x: int) -> int:
        return bad_func()

    assert f(3) == 6
    with pytest.raises(deal.PreContractError) as exc_info:
        f(4)
    assert exc_info.value.source == 'False'


@pytest.mark.filterwarnings('ignore:It is pytest but deal is disabled')
def test_dispatch_works_with_disabled_contracts():
    state.disable()
    try:
        f = deal.dispatch(double)
        f.register(double3)
        f.register(double4)
        assert f(4) == 8
    finally:
        state.enable()



from __future__ import annotations

from typing import Iterator


ERROR_FORMAT = 'DEL{code:03d}'


class Error:
    __slots__ = ('row', 'col', 'code', 'text', 'value')

    def __init__(
        self, *,
        row: int,
        col: int,
        code: int,
        text: str,
        value: str | None = None,
    ) -> None:
        self.row = row
        self.col = col
        self.code = code
        self.text = text
        self.value = value

    @property
    def full_code(self) -> str:
        return ERROR_FORMAT.format(code=self.code)

    @property
    def message(self) -> str:
        msg = self.full_code + ' ' + self.text
        if self.value:
            msg += f' ({self.value})'
        return msg

    def __iter__(self) -> Iterator[int | str]:
        yield self.row
        yield self.col
        yield self.message

    def __str__(self) -> str:
        return self.message

    def __repr__(self) -> str:
        name = type(self).__name__
        return f'{name}(row={self.row}, col={self.col}, code={self.code})'

    def __hash__(self) -> int:
        return hash((self.row, self.col, self.code))



from __future__ import annotations

import ast
import json
from argparse import ArgumentParser
from pathlib import Path
from textwrap import dedent, indent
from typing import Iterable, Iterator

from .._colors import get_colors, highlight
from ..linter import Checker
from ._base import Command
from ._common import get_paths


TEMPLATE = '  {blue}{row}{end}:{blue}{col}{end} {magenta}{code}{end} {yellow}{text}{end}'
VALUE = ' {magenta}({value}){end}'
POINTER = '{magenta}^{end}'


class LintCommand(Command):
    """Run linter against the given files.

    ```bash
    python3 -m deal lint project/
    ```

    Options:

    + `--json`: output violations as [json per line][ndjson].
    + `--nocolor`: output violations in human-friendly format but without colors.
      Useful for running linter on CI.

    Exit code is equal to the found violations count.
    See [linter][linter] documentation for more details.

    [ndjson]: http://ndjson.org/
    [linter]: https://deal.readthedocs.io/basic/linter.html
    """

    @staticmethod
    def init_parser(parser: ArgumentParser) -> None:
        parser.add_argument('--json', action='store_true', help='json output')
        parser.add_argument('--nocolor', action='store_true', help='colorless output')
        parser.add_argument('paths', nargs='*', default='.')

    def __call__(self, args) -> int:
        prev = None
        errors = list(self.get_errors(paths=args.paths))
        colors = get_colors(args)
        for error in errors:
            if args.json:
                self.print(json.dumps(error))
                continue

            # print file path
            if error['path'] != prev:
                self.print('{green}{path}{end}'.format(**colors, **error))
            prev = error['path']

            # print message
            line = TEMPLATE.format(**colors, **error)
            if error['value']:
                line += VALUE.format(**colors, **error)
            self.print(line)

            # print code line
            pointer = ' ' * error['col'] + POINTER.format(**colors)
            content = error['content']
            if not args.nocolor:
                content = highlight(content)
            content += '\n' + pointer
            content = indent(dedent(content), prefix='    ')
            self.print(content)
        return len(errors)

    @staticmethod
    def get_errors(paths: Iterable[str]) -> Iterator[dict]:
        for arg in paths:
            for path in get_paths(Path(arg)):
                content = path.read_text()
                checker = Checker(
                    filename=str(path),
                    tree=ast.parse(content),
                )
                lines = content.split('\n')
                for error in checker.get_errors():
                    yield dict(
                        path=str(path),
                        row=error.row,
                        col=error.col,
                        code=error.full_code,
                        text=error.text,
                        value=error.value,
                        content=lines[error.row - 1],
                    )



from __future__ import annotations

import pytest

import deal


try:
    import vaa
except ImportError:
    vaa = None

MarshMallowScheme: object | None
if vaa is None:
    MarshMallowScheme = None
else:
    import marshmallow

    @vaa.marshmallow
    class MarshMallowScheme(marshmallow.Schema):
        name = marshmallow.fields.Str()
        kwargs = marshmallow.fields.Dict(required=False)


class CustomScheme(deal.Scheme):
    def is_valid(self) -> bool:
        if not isinstance(self.data['name'], str):
            self.errors = vaa.Error.parse({'name': ['Not a valid string.']})
            return False
        return True


if MarshMallowScheme:
    SCHEMES = (MarshMallowScheme, CustomScheme)
else:
    SCHEMES = (CustomScheme, )


@pytest.mark.skipif(vaa is None, reason='vaa is not installed')
@pytest.mark.parametrize('scheme', SCHEMES)
def test_scheme_string_validation_args_correct(scheme):
    @deal.pre(scheme)
    def func(name):
        return name * 2

    assert func('Chris') == 'ChrisChris'

    with pytest.raises(deal.PreContractError):
        func(123)

    try:
        func(123)
    except deal.PreContractError as e:
        assert e.args[0] == [vaa.Error(field='name', message='Not a valid string.')]


@pytest.mark.skipif(vaa is None, reason='vaa is not installed')
@pytest.mark.parametrize('scheme', SCHEMES)
def test_method_chain_decorator_with_scheme_is_fulfilled(scheme):
    @deal.pre(scheme)
    @deal.pre(lambda name: name != 'Oleg')
    def func(name):
        return name * 2

    assert func('Chris') == 'ChrisChris'

    with pytest.raises(deal.PreContractError):
        func(123)

    with pytest.raises(deal.PreContractError):
        func('Oleg')


@pytest.mark.skipif(vaa is None, reason='vaa is not installed')
@pytest.mark.parametrize('scheme', SCHEMES)
def test_scheme_contract_is_satisfied_when_setting_arg(scheme):
    @deal.inv(scheme)
    class User:
        name: str | int = ''

    user = User()

    user.name = 'Chris'

    with pytest.raises(deal.InvContractError):
        user.name = 123

    try:
        user.name = 123
    except deal.InvContractError as e:
        assert e.args[0] == [vaa.Error(field='name', message='Not a valid string.')]


@pytest.mark.skipif(vaa is None, reason='vaa is not installed')
@pytest.mark.parametrize('scheme', SCHEMES)
def test_scheme_contract_is_satisfied_within_chain(scheme):
    @deal.inv(lambda user: user.name != 'Oleg')
    @deal.inv(scheme)
    @deal.inv(lambda user: user.name != 'Chris')
    class User:
        name: str | int = ''

    user = User()
    user.name = 'Gram'

    user = User()
    with pytest.raises(deal.InvContractError):
        user.name = 'Oleg'

    user = User()
    with pytest.raises(deal.InvContractError):
        user.name = 123

    user = User()
    with pytest.raises(deal.InvContractError):
        user.name = 'Chris'


@pytest.mark.skipif(vaa is None, reason='vaa is not installed')
@pytest.mark.parametrize('scheme', SCHEMES)
def test_scheme_contract_is_satisfied_when_passing_args(scheme):
    @deal.pre(scheme)
    def func1(name):
        return name * 2

    assert func1('Chris') == 'ChrisChris'
    assert func1(name='Chris') == 'ChrisChris'

    @deal.pre(scheme)
    def func2(**kwargs):
        return kwargs['name'] * 3

    assert func2(name='Chris') == 'ChrisChrisChris'

    @deal.pre(scheme)
    def func3(name='Max'):
        return name * 2

    assert func3() == 'MaxMax'


@pytest.mark.skipif(vaa is None, reason='vaa is not installed')
@pytest.mark.parametrize('scheme', SCHEMES)
def test_scheme_errors_rewrite_message(scheme):
    @deal.pre(scheme, message='old message')
    def func(name):
        return name * 2

    try:
        func(2)
    except deal.PreContractError as exc:
        assert exc.errors == [vaa.Error(field='name', message='Not a valid string.')]
    else:
        raise AssertionError('exception was not raised')


def test_underscore_validator():
    @deal.pre(lambda _: _.a != _.b, message='actual message')
    def func(a, b=1):
        return a + b

    func(2)
    func(1, 3)
    func(a=1, b=3)
    with pytest.raises(deal.PreContractError) as exc_info:
        func(1)
    assert exc_info.value.args == ('actual message',)


def test_underscore_validator_default_message():
    @deal.pre(lambda _: _.a != _.b)
    def func(a, b=1):
        return a + b

    func(2)
    func(1, 3)
    func(a=1, b=3)
    with pytest.raises(deal.PreContractError) as exc_info:
        func(1)
    assert exc_info.value.args == tuple()


@pytest.mark.skipif(vaa is None, reason='vaa is not installed')
def test_default_error():
    """
    If no error provided by the validator, return the default one.
    """
    class CustomScheme(deal.Scheme):
        def is_valid(self) -> bool:
            return isinstance(self.data['name'], str)

    @deal.pre(CustomScheme, message='oh hi mark')
    def func(name):
        return name * 2
    func('hi')
    with pytest.raises(deal.PreContractError) as exc_info:
        func(123)
    assert exc_info.value.message == 'oh hi mark'



from __future__ import annotations

import ast
from collections import deque
from contextlib import suppress
from functools import partial
from pathlib import Path
from typing import Callable, Iterator, NamedTuple, Tuple, Type, TypeVar

from .._stub import EXTENSION, StubFile, StubsManager


try:
    import astroid
except ImportError:
    astroid = None


T = TypeVar('T')
N = Tuple[Type[T], Type[T]]
Handler = Callable[..., 'Token | Iterator[Token] | None']


def _get_type(name: str) -> tuple[type, type]:
    ast_type = getattr(ast, name)
    if astroid is not None:         # pragma: has-astroid
        astroid_type = getattr(astroid, name)
        return (ast_type, astroid_type)
    return (ast_type, ast_type)     # pragma: no-astroid


class TOKENS:
    ASSERT: N[ast.Assert] = _get_type('Assert')
    ATTR: N[ast.Attribute] = _get_type('Attribute')
    BIN_OP: N[ast.BinOp] = _get_type('BinOp')
    CALL: N[ast.Call] = _get_type('Call')
    COMPARE: N[ast.Compare] = _get_type('Compare')
    GLOBAL: N[ast.Global] = _get_type('Global')
    LAMBDA: N[ast.Lambda] = _get_type('Lambda')
    NONLOCAL: N[ast.Nonlocal] = _get_type('Nonlocal')
    RAISE: N[ast.Raise] = _get_type('Raise')
    RETURN: N[ast.Return] = _get_type('Return')
    YIELD_FROM: N[ast.YieldFrom] = _get_type('YieldFrom')
    YIELD: N[ast.Yield] = _get_type('Yield')


DEFAULT_LINE = 0
DEFAULT_COL = 1


class Token(NamedTuple):
    line: int = DEFAULT_LINE
    col: int = DEFAULT_COL
    value: object | None = None
    marker: str | None = None  # marker name or error message


def traverse(body: list[ast.AST | astroid.NodeNG]) -> Iterator[ast.AST | astroid.NodeNG]:
    for expr in body:
        if isinstance(expr, ast.AST):
            yield from _traverse_ast(expr)
        else:
            yield from _traverse_astroid(expr)


def _traverse_ast(node: ast.AST) -> Iterator[ast.AST]:
    todo = deque([node])
    while todo:
        node = todo.popleft()
        if isinstance(node, ast.Try):
            for h in node.handlers:
                todo.extend(h.body)
            todo.extend(node.orelse)
            todo.extend(node.finalbody)
        else:
            todo.extend(ast.iter_child_nodes(node))
            yield node


def _traverse_astroid(node: astroid.NodeNG) -> Iterator[astroid.NodeNG]:
    todo: deque[astroid.NodeNG] = deque([node])
    while todo:
        node = todo.popleft()
        if isinstance(node, astroid.Try):
            for h in node.handlers:
                todo.extend(h.body)
            todo.extend(node.orelse)
            todo.extend(node.finalbody)
        else:
            todo.extend(node.get_children())
            yield node


def get_name(expr: ast.AST | astroid.NodeNG) -> str | None:
    if isinstance(expr, ast.Name):
        return expr.id
    if astroid is not None and isinstance(expr, astroid.Name):
        return expr.name

    if astroid is not None and isinstance(expr, astroid.Attribute):
        left = get_name(expr.expr)
        if left is None:
            return None
        return left + '.' + expr.attrname
    if isinstance(expr, ast.Attribute):
        left = get_name(expr.value)
        if left is None:
            return None
        return left + '.' + expr.attr

    return None


def get_full_name(
    expr: astroid.Module | astroid.FunctionDef | astroid.UnboundMethod | astroid.ClassDef,
) -> Tuple[str, str]:
    if expr.parent is None:
        return '', expr.name

    if isinstance(expr.parent, astroid.Module):
        return expr.parent.qname(), expr.name

    if isinstance(expr.parent, (astroid.FunctionDef, astroid.UnboundMethod)):
        module_name, func_name = get_full_name(expr=expr.parent)
        return module_name, func_name + '.' + expr.name

    if isinstance(expr.parent, astroid.ClassDef):
        module_name, class_name = get_full_name(expr=expr.parent)
        return module_name, class_name + '.' + expr.name

    path, _, func_name = expr.qname().rpartition('.')
    return path, func_name


def infer(expr: ast.AST | astroid.NodeNG) -> Tuple[astroid.NodeNG, ...]:
    if isinstance(expr, ast.AST):
        return tuple()
    with suppress(astroid.InferenceError, RecursionError):
        guesses = expr.infer()
        if guesses is astroid.Uninferable:  # pragma: no cover
            return tuple()
        return tuple(g for g in guesses if repr(g) != 'Uninferable')
    return tuple()


def get_stub(
    module_name: str | None,
    expr: astroid.FunctionDef,
    stubs: StubsManager,
) -> StubFile | None:
    if not module_name:
        return None
    stub = stubs.get(module_name)
    if stub is not None:
        return stub

    module = _get_module(expr=expr)
    if module is None or module.file is None:
        return None  # pragma: no coverage
    path = Path(module.file).with_suffix(EXTENSION)
    if not path.exists():
        return None
    return stubs.read(path=path)


def _get_module(expr: astroid.NodeNG) -> astroid.Module | None:
    if isinstance(expr, astroid.Module):
        return expr
    if expr.parent is None:
        return None
    return _get_module(expr.parent)


class Extractor:
    __slots__ = ('handlers', )
    handlers: dict[type, Handler]

    def __init__(self) -> None:
        self.handlers = dict()

    def register(
        self, *types: type | Callable[[], type],
    ) -> Callable[[Handler], Handler]:
        types = tuple(self._eval_types(types))
        return partial(self._register, types)

    @staticmethod
    def _eval_types(types: tuple[type | Callable[[], type], ...]) -> Iterator[type]:
        for t in types:
            if isinstance(t, type):
                yield t
                continue
            try:
                yield t()
            except (NameError, AttributeError):  # pragma: no-astroid
                pass

    def _register(self, types: Tuple[type, ...], handler: Handler) -> Handler:
        for tp in types:
            # it's here to have `getattr` to get nodes from `ast` module
            # that are available only in some Python versions.
            if tp is None:
                continue  # pragma: no coverage
            self.handlers[tp] = handler
        return handler

    def __call__(self, body: list, **kwargs) -> Iterator[Token]:
        for expr in traverse(body=body):
            for token in self._handle(expr=expr, **kwargs):
                yield self._ensure_node_info(expr=expr, token=token)

    def _handle(self, expr: ast.AST | astroid.NodeNG, **kwargs) -> Iterator[Token]:
        handler = self.handlers.get(type(expr))
        if handler is None:
            return
        token = handler(expr=expr, **kwargs)
        if token is None:
            return
        if isinstance(token, Token):
            yield token
            return
        yield from token

    @staticmethod
    def _ensure_node_info(token: Token, expr: ast.AST | astroid.NodeNG) -> Token:
        if isinstance(expr, ast.AST):
            if not isinstance(expr, (ast.stmt, ast.expr)):  # pragma: no cover
                return token
        if token.line == DEFAULT_LINE:
            token = token._replace(line=expr.lineno)
        if token.col == DEFAULT_COL:
            token = token._replace(col=expr.col_offset)
        return token



from __future__ import annotations

import inspect
from functools import lru_cache
from typing import TYPE_CHECKING, Any, Callable

from .._exceptions import ContractError
from .._types import ExceptionType


try:
    import vaa
except ImportError:
    vaa = None


if TYPE_CHECKING:
    Args = tuple[object, ...]
    Kwargs = dict[str, object]


@lru_cache(maxsize=16)
def _get_signature(function: Callable) -> inspect.Signature:
    return inspect.signature(function)


def _args_to_vars(
    *,
    args: Args,
    kwargs: Kwargs,
    signature: inspect.Signature | None,
    keep_result: bool = True,
) -> dict[str, object]:
    """Convert args and kwargs into dict of params based on the given function.

    For simple validators the validator is passed as function.
    """
    if signature is None:
        return kwargs

    params = kwargs.copy()
    # Do not pass argument named `result` into the function.
    # It is a hack for `deal.ensure` with `vaa` validator.
    if not keep_result and 'result' in kwargs:
        kwargs = kwargs.copy()
        del kwargs['result']

    # assign *args to real names
    for name, param in signature.parameters.items():
        params[name] = param.default
    params.update(signature.bind(*args, **kwargs).arguments)
    return params


class AttrDict(dict):
    __slots__ = ()

    def __getattr__(self, name: str):
        return self[name]


class Validator:
    __slots__ = (
        'exception',
        'signature',
        'validate',
        'validator',
        'raw_validator',
        'message',
        'function',
    )

    exception: ExceptionType
    signature: inspect.Signature | None
    validate: Any
    validator: Any
    raw_validator: Any
    message: str | None
    function: Any

    def __init__(
        self,
        validator, *,
        message: str | None = None,
        exception: ExceptionType,
    ) -> None:
        self.validate = self._init
        self.raw_validator = validator
        self.message = message
        self.exception = exception
        self.function = None
        if message and isinstance(self.exception, type):
            self.exception = self.exception(message)

    @property
    def exception_type(self) -> type[Exception]:
        if isinstance(self.exception, Exception):
            return type(self.exception)
        return self.exception

    def _exception(
        self, *,
        message: str | None = None,
        errors: Kwargs | None = None,
        params: Kwargs | None = None,
    ) -> Exception:
        exception = self.exception
        if isinstance(exception, Exception):
            if not message and exception.args:
                message = exception.args[0]
            exception = type(exception)

        # raise beautiful ContractError
        if issubclass(exception, ContractError):
            return exception(
                message=message or '',
                validator=self.validator,
                errors=errors,
                params=params,
                origin=getattr(self, 'function', None),
            )

        # raise boring custom exception
        args: list[Any] = []
        if message:
            args.append(message)
        if errors:
            args.append(errors)
        return exception(*args)

    def _wrap_vaa(self) -> Any | None:
        if vaa is None:  # pragma: no cover
            return None
        try:
            return vaa.wrap(self.raw_validator, simple=False)
        except TypeError:
            pass
        if hasattr(self.raw_validator, 'is_valid'):
            return self.raw_validator
        return None

    def init(self) -> None:
        # implicitly wrap in vaa.simple only funcs with one `_` argument.
        self.signature = None
        val_signature = _get_signature(self.raw_validator)

        # validator with a short signature
        if set(val_signature.parameters) == {'_'}:
            self.validator = self.raw_validator
            self.validate = self._short_validation
            if self.function is not None:
                self.signature = _get_signature(self.function)
            return

        vaa_validator = self._wrap_vaa()
        if vaa_validator is None:
            # vaa validator
            self.validator = self.raw_validator
            self.validate = self._explicit_validation
            self.signature = val_signature
        else:
            # validator with the same signature as the function
            self.validator = vaa_validator
            self.validate = self._vaa_validation
            if self.function is not None:
                self.signature = _get_signature(self.function)

    def _init(self, args: Args, kwargs: Kwargs, exc=None) -> None:
        """
        Called as `validator` when the function is called in the first time.
        Does some costly deferred initializations (involving `inspect`).
        Then sets more appropriate validator as `validator` and calls it.
        """
        self.init()
        self.validate(args, kwargs, exc)

    def _vaa_validation(self, args: Args, kwargs: Kwargs, exc=None) -> None:
        """Validate contract using vaa wrapped validator.
        """

        # if it is a decorator for a function, convert positional args into named ones.
        params = _args_to_vars(
            args=args,
            kwargs=kwargs,
            signature=self.signature,
            keep_result=False,
        )

        # validate
        validator = self.validator(data=params)
        if validator.is_valid():
            return

        # if no errors returned, raise the default exception
        errors = validator.errors
        if not errors:
            raise self._exception(params=params) from exc

        raise self._exception(errors=errors, params=params) from exc

    def _explicit_validation(self, args: Args, kwargs: Kwargs, exc=None) -> None:
        """Validate contract using explicit validator.

        Explicit validator is a function that has the same signature
        as the wrapped function.
        """
        validation_result = self.validator(*args, **kwargs)
        # is invalid (validator returns error message)
        if type(validation_result) is str:
            params = _args_to_vars(args=args, kwargs=kwargs, signature=self.signature)
            raise self._exception(message=validation_result, params=params) from exc
        # is valid (truely result)
        if validation_result:
            return
        # is invalid (falsy result)
        params = _args_to_vars(args=args, kwargs=kwargs, signature=self.signature)
        raise self._exception(params=params) from exc

    def _short_validation(self, args: Args, kwargs: Kwargs, exc=None) -> None:
        """Validate contract using short validator.

        Short validator is a function that has a short signature
        (accepts only one `_` argument).
        """
        params = _args_to_vars(
            args=args,
            kwargs=kwargs,
            signature=self.signature,
            keep_result=False,
        )
        validation_result = self.validator(AttrDict(params))
        # is invalid (validator returns error message)
        if type(validation_result) is str:
            raise self._exception(message=validation_result, params=params) from exc
        # is valid (truely result)
        if validation_result:
            return
        # is invalid (falsy result)
        raise self._exception(params=params) from exc


class RaisesValidator(Validator):
    __slots__ = ('exceptions', )
    exceptions: tuple[type[Exception]]

    def __init__(self, exceptions, exception, message) -> None:
        self.exceptions = exceptions
        self.validator = None
        super().__init__(validator=None, message=message, exception=exception)

    def init(self) -> None:
        self.signature = _get_signature(self.function)
        self.validate = self._validate

    def _init(self, args: Args, kwargs: Kwargs, exc=None) -> None:
        self.init()
        self.validate(args, kwargs, exc=exc)

    def _validate(self, args: Args, kwargs: Kwargs, exc: Exception | None = None) -> None:
        assert exc is not None
        exc_type = type(exc)
        if exc_type in self.exceptions:
            return
        raise self._exception() from exc_type


class ReasonValidator(Validator):
    __slots__ = ('event', )
    event: type[Exception]

    def __init__(self, event, **kwargs) -> None:
        self.event = event
        super().__init__(**kwargs)


class InvariantValidator(Validator):
    def _vaa_validation(self, args: Args, kwargs: Kwargs, exc=None) -> None:
        return super()._vaa_validation((), vars(args[0]), exc=exc)

    def _short_validation(self, args: Args, kwargs: Kwargs, exc=None) -> None:
        return super()._short_validation((), vars(args[0]), exc=exc)



from __future__ import annotations

from functools import partial, update_wrapper
from types import MethodType
from typing import Callable, TypeVar

from ._validators import InvariantValidator


T = TypeVar('T', bound=type)
ATTR = '_deal_invariants'
DEAL_ATTRS = frozenset({
    '_deal_patched_method',
    '_deal_validate',
    ATTR,
})


class InvariantedClass:
    _deal_invariants: list[InvariantValidator]

    def _deal_validate(self) -> None:
        for validator in self._deal_invariants:
            validator.validate((self,), {})

    def _deal_patched_method(self, method: Callable, *args, **kwargs):
        self._deal_validate()
        result = method(*args, **kwargs)
        self._deal_validate()
        return result

    def __getattribute__(self, name: str):
        attr = super().__getattribute__(name)
        if name in DEAL_ATTRS:
            return attr
        if not isinstance(attr, MethodType):
            return attr
        patched_method = partial(self._deal_patched_method, attr)
        return update_wrapper(patched_method, attr)

    def __setattr__(self, name: str, value) -> None:
        super().__setattr__(name, value)
        self._deal_validate()


def invariant(validator: InvariantValidator, _class: T) -> T:
    invs = getattr(_class, ATTR, None)
    if invs is None:
        patched_class = type(
            _class.__name__ + 'Invarianted',
            (InvariantedClass, _class),
            {ATTR: [validator]},
        )
    else:
        patched_class = type(
            _class.__name__,
            (_class, ),
            {ATTR: invs + [validator]},
        )
    return patched_class  # type: ignore[return-value]



from __future__ import annotations

from enum import Enum
from pathlib import Path
from typing import Iterator, NamedTuple, Union

from ._contract import Category
from ._extractors import get_value
from ._func import Func
from ._rules import CheckMarkers, CheckRaises


try:
    import astroid
except ImportError:
    astroid = None

Priority = int


class TransformationType(Enum):
    RAISES = 'raises'
    HAS = 'has'
    SAFE = 'safe'
    PURE = 'pure'
    IMPORT = 'import'


class AppendText(NamedTuple):
    line: int
    text: str

    def apply(self, lines: list[str]) -> None:
        content = lines[self.line - 1]
        content = content.rstrip('\n')
        content += f'{self.text}\n'
        lines[self.line - 1] = content

    @property
    def key(self) -> tuple[int, Priority]:
        return (self.line, 1)


class InsertText(NamedTuple):
    line: int
    text: str

    def apply(self, lines: list[str]) -> None:
        lines.insert(self.line - 1, f'{self.text}\n')

    @property
    def key(self) -> tuple[int, Priority]:
        return (self.line, 2)


class InsertContract(NamedTuple):
    line: int
    contract: Category
    args: list[str]
    indent: int

    def apply(self, lines: list[str]) -> None:
        lines.insert(self.line - 1, f'{self}\n')

    @property
    def key(self) -> tuple[int, Priority]:
        return (self.line, 3)

    def __str__(self) -> str:
        args = ', '.join(self.args)
        if not self.args and self.contract.brackets_optional:
            dec = f'@deal.{self.contract.value}'
        else:
            dec = f'@deal.{self.contract.value}({args})'
        return ' ' * self.indent + dec


class Remove(NamedTuple):
    line: int

    def apply(self, lines: list[str]) -> None:
        lines.pop(self.line - 1)

    @property
    def key(self) -> tuple[int, Priority]:
        return (self.line, 4)


Mutation = Union[AppendText, InsertText, InsertContract, Remove]


class Transformer(NamedTuple):
    """Transformer adds deal decorators into the given script.
    """
    content: str
    path: Path
    types: set[TransformationType]
    mutations: list[Mutation] = []
    quote: str = "'"

    def transform(self) -> str:
        if astroid is None:  # pragma: no-astroid
            raise ImportError('astroid is required for generating stubs')
        self.mutations.clear()
        tree = astroid.parse(self.content, path=self.path)
        for func in Func.from_astroid(tree):
            self._collect_mutations(func)
        self.mutations.extend(self._mutations_pure())
        self.mutations.extend(self._mutations_import(tree))
        return self._apply_mutations(self.content)

    def _collect_mutations(self, func: Func) -> None:
        self.mutations.extend(self._mutations_excs(func))
        self.mutations.extend(self._mutations_markers(func))
        self.mutations.extend(self._mutations_property(func))

    def _mutations_excs(self, func: Func) -> Iterator[Mutation]:
        """Add @deal.raises or @deal.safe if needed.
        """
        cats = {Category.RAISES, Category.SAFE, Category.PURE}

        # collect declared exceptions
        declared: list[Union[str, type]] = []
        for contract in func.contracts:
            if contract.category not in cats:
                continue
            declared.extend(contract.exceptions)

        # collect undeclared exceptions
        excs: set[str] = set()
        for error in CheckRaises().get_undeclared(func, declared):
            assert isinstance(error.value, str)
            excs.add(error.value)

        # if no new exceptions found, add deal.safe
        if not excs:
            if declared:
                return
            if self._disabled(TransformationType.SAFE, TransformationType.PURE):
                return
            if func.has_contract(Category.PURE, Category.SAFE):
                return
            yield InsertContract(
                line=self._get_insert_line(func),
                indent=func.col,
                contract=Category.SAFE,
                args=[],
            )
            return

        # if new exceptions detected, remove old contracts and add a new deal.raises
        if self._disabled(TransformationType.RAISES):
            return
        for contract in func.contracts:
            if contract.category not in cats:
                continue
            yield Remove(contract.line)
            if contract.category == Category.PURE:
                yield InsertContract(
                    line=self._get_insert_line(func),
                    indent=func.col,
                    contract=Category.HAS,
                    args=[],
                )
        contract_args = [self._exc_as_str(exc) for exc in declared]
        contract_args.extend(sorted(excs))
        yield InsertContract(
            line=self._get_insert_line(func),
            indent=func.col,
            contract=Category.RAISES,
            args=contract_args,
        )

    @staticmethod
    def _exc_as_str(exc) -> str:
        if isinstance(exc, str):
            return exc
        return exc.__name__

    def _mutations_markers(self, func: Func) -> Iterator[Mutation]:
        """Add @deal.has if needed.
        """
        if self._disabled(TransformationType.HAS, TransformationType.PURE):
            return
        cats = {Category.HAS, Category.PURE}

        # collect declared markers
        declared: list[str] = []
        for contract in func.contracts:
            if contract.category not in cats:
                continue
            for arg in contract.args:
                value = get_value(arg)
                if isinstance(value, str):
                    declared.append(value)

        # collect undeclared markers
        markers: set[str] = set()
        for error in CheckMarkers().get_undeclared(func, set(declared)):
            assert isinstance(error.value, str)
            markers.add(error.value)

        # if no new markers found, add deal.has()
        if not markers:
            if func.has_contract(Category.PURE, Category.HAS):
                return
            yield InsertContract(
                line=self._get_insert_line(func),
                indent=func.col,
                contract=Category.HAS,
                args=[],
            )
            return

        # if new markers detected, remove old contracts and add a new deal.has
        for contract in func.contracts:
            if contract.category not in cats:
                continue
            yield Remove(contract.line)
            if contract.category == Category.PURE:
                yield InsertContract(
                    line=self._get_insert_line(func),
                    indent=func.col,
                    contract=Category.SAFE,
                    args=[],
                )
        contract_args = [self._exc_as_str(marker) for marker in declared]
        contract_args.extend(sorted(markers))
        yield InsertContract(
            line=self._get_insert_line(func),
            indent=func.col,
            contract=Category.HAS,
            args=[f'{self.quote}{arg}{self.quote}' for arg in contract_args],
        )

    def _mutations_property(self, func: Func) -> Iterator[Mutation]:
        assert isinstance(func.node, astroid.FunctionDef)
        if func.node.decorators is None:
            return
        assert isinstance(func.node.decorators, astroid.Decorators)
        for decorator in func.node.decorators.nodes:
            if not isinstance(decorator, astroid.Name):
                continue
            if decorator.name not in {'property', 'cached_property'}:
                continue
            if not self._has_mutation_on_line(decorator.lineno + 1):
                continue
            yield AppendText(decorator.lineno, '  # type: ignore[misc]')

    def _has_mutation_on_line(self, line: int) -> bool:
        return any(mutation.line == line for mutation in self.mutations)

    def _mutations_import(self, tree: astroid.Module) -> Iterator[Mutation]:
        """Add `import deal` if needed.
        """
        if self._disabled(TransformationType.IMPORT):
            return
        if not self.mutations:
            return
        # check if already imported
        for stmt in tree.body:
            if not isinstance(stmt, astroid.Import):
                continue
            for name, _ in stmt.names:
                if name == 'deal':
                    return

        # We insert the import after `__future__` imports and module imports.
        # We don't skip `from` imports, though, because they can be multiline.
        line = 1
        for stmt in tree.body:
            if isinstance(stmt, astroid.Import):
                line = stmt.lineno + 1
            if isinstance(stmt, astroid.ImportFrom):
                if stmt.modname == '__future__':
                    line = stmt.lineno + 1
        yield InsertText(line=line, text='import deal')

    def _mutations_pure(self) -> Iterator[Mutation]:
        """Merge `@deal.safe` and `@deal.has` into `@deal.pure` if needed.
        """
        if self._disabled(TransformationType.PURE):
            return
        if not self.mutations:
            return

        # find lines that have both @deal.has and @deal.safe
        lines_has = set()
        lines_safe = set()
        for mut in self.mutations:
            if not isinstance(mut, InsertContract):
                continue
            if mut.contract == Category.HAS and mut.args == []:
                lines_has.add(mut.line)
            if mut.contract == Category.SAFE:
                lines_safe.add(mut.line)
        lines = lines_safe & lines_has

        # remove @deal.has and @deal.safe mutations, replace by @deal.pure
        old_cats = {Category.HAS, Category.SAFE}
        for mut in self.mutations.copy():
            if not isinstance(mut, InsertContract):
                continue
            if mut.line not in lines:
                continue
            assert mut.contract in old_cats, 'unexpected contract generated'
            self.mutations.remove(mut)
            if mut.contract == Category.SAFE:
                yield mut._replace(contract=Category.PURE)

        # If PURE tranformation is enabled,
        # we emit @deal.safe and @deal.has even if they are disabled, so they can be
        # merged into @deal.pure. So, if they are disabled and were not merged,
        # drop them here.
        if self._disabled(TransformationType.SAFE):
            for mut in self.mutations.copy():
                if isinstance(mut, InsertContract) and mut.contract == Category.SAFE:
                    self.mutations.remove(mut)
        if self._disabled(TransformationType.HAS):
            for mut in self.mutations.copy():
                if isinstance(mut, InsertContract) and mut.contract == Category.HAS:
                    self.mutations.remove(mut)

    @staticmethod
    def _get_insert_line(func: Func) -> int:
        assert isinstance(func.node, astroid.FunctionDef)
        line = func.line
        if func.node.decorators is None:
            return line
        assert isinstance(func.node.decorators, astroid.Decorators)
        for decorator in func.node.decorators.nodes:
            # some Python versions point to the first decorator, some to `def`
            if decorator.lineno < func.line:
                return func.line  # pragma: no cover
            if not isinstance(decorator, astroid.Name):
                continue
            if decorator.name in {'staticmethod', 'classmethod'}:
                continue
            line = decorator.lineno + 1
        return line

    def _apply_mutations(self, content: str) -> str:
        if not self.mutations:
            return content
        lines = content.splitlines(keepends=True)
        self.mutations.sort(key=lambda x: x.key, reverse=True)
        for mutation in self.mutations:
            mutation.apply(lines)
        return ''.join(lines)

    def _disabled(self, *expected: TransformationType) -> bool:
        """Check if all of the given types are disabled.
        """
        return not bool(self.types & set(expected))



from __future__ import annotations

from functools import update_wrapper
from typing import TYPE_CHECKING, Callable, Generic, TypeVar

from .._exceptions import NoMatchError, PreContractError
from .._state import state


if TYPE_CHECKING:
    from ._contracts import Contracts


F = TypeVar('F', bound=Callable)
ATTR = '__deal_contract'


class Dispatch(Generic[F]):
    _functions: list[F]
    __call__: F

    def __init__(self) -> None:
        self._functions = []

    @classmethod
    def wrap(cls, func: F) -> Dispatch[F]:
        self = cls()
        update_wrapper(wrapper=self, wrapped=func)
        return self

    def register(self, function: F) -> F:
        self._functions.append(function)
        return function

    def __call__(self, *args, **kwargs):  # type: ignore[no-redef]
        exceptions = []
        contracts: Contracts | None
        old_state = state.debug
        state.debug = True
        try:
            for func in self._functions:
                try:
                    return func(*args, **kwargs)
                except PreContractError as exc:
                    contracts = getattr(func, ATTR, None)
                    if contracts is None:
                        raise
                    if exc.origin is not contracts.func:
                        raise
                    exceptions.append(exc)
        finally:
            state.debug = old_state
        raise NoMatchError(tuple(exceptions))



from io import StringIO
from pathlib import Path
from textwrap import dedent

import pytest

from deal._cli import main


try:
    import astroid
except ImportError:
    astroid = None


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
@pytest.mark.parametrize('flags, given, expected', [
    (
        [],
        """
            @deal.post(lambda x: x > 0)
            def f(x):
                print(1/0)
                return -1
        """,
        """
            import deal

            @deal.has('stdout')
            @deal.raises(ZeroDivisionError)
            @deal.post(lambda x: x > 0)
            def f(x):
                print(1/0)
                return -1
        """,
    ),
    (
        ['--types', 'raises', 'safe'],
        """
            import deal
            @deal.post(lambda x: x > 0)
            def f(x):
                print(1/0)
                return -1
        """,
        """
            import deal
            @deal.raises(ZeroDivisionError)
            @deal.post(lambda x: x > 0)
            def f(x):
                print(1/0)
                return -1
        """,
    ),
    (
        ['--types', 'has', '--double-quotes'],
        """
            import deal
            @deal.post(lambda x: x > 0)
            def f(x):
                print(1/0)
                return -1
        """,
        """
            import deal
            @deal.has("stdout")
            @deal.post(lambda x: x > 0)
            def f(x):
                print(1/0)
                return -1
        """,
    ),
    (
        [],
        """
            import deal
            @deal.pure
            def f(x):
                return x
        """,
        """
            import deal
            @deal.pure
            def f(x):
                return x
        """,
    ),
])
def test_decorate_command(flags: list, given: str, expected: str, tmp_path: Path):
    file_path = tmp_path / 'example.py'
    file_path.write_text(dedent(given))
    stream = StringIO()
    code = main(['decorate', *flags, '--', str(tmp_path)], stream=stream)
    assert code == 0

    stream.seek(0)
    captured = stream.read()
    assert str(file_path) in captured
    assert file_path.read_text().lstrip('\n') == dedent(expected).lstrip('\n')



from __future__ import annotations

import os
import warnings
from types import MappingProxyType
from typing import Callable, Mapping, TypeVar


T = TypeVar('T', bound=Callable)
PERMAMENT_ERROR = RuntimeError('contracts are permanently disabled')
PROD_ENV = MappingProxyType(dict(
    LAMBDA_TASK_ROOT='AWS',
    GCLOUD_PROJECT='GCP',
))
TEST_ENV = MappingProxyType(dict(
    PYTEST_CURRENT_TEST='pytest',
    CI='CI',
))


class _State:
    __slots__ = ('debug', 'removed', 'color')
    debug: bool
    removed: bool
    color: bool

    def __init__(self) -> None:
        self.removed = False
        self.reset()

    def reset(self) -> None:
        """Restore contracts state to the default.

        All contracts are disabled on production by default.
        See [runtime][runtime] documentation.

        [runtime]: https://deal.readthedocs.io/basic/runtime.html
        """
        if self.removed:
            raise PERMAMENT_ERROR
        self.debug = __debug__
        self.color = 'NO_COLOR' not in os.environ

    def enable(self, warn: bool = True) -> None:
        """Enable all contracts.

        By default, deal will do a few sanity checks to make sure you haven't
        unintentionally enabled contracts on a production environment.
        Pass `warn=False` to disable this behavior.
        """
        if self.removed:
            raise PERMAMENT_ERROR
        self.debug = True
        if warn:
            # coverage freaks out here on Python 3.10
            if not __debug__:  # pragma: no cover
                msg = 'It is production but deal is enabled. Is it intentional?'
                warnings.warn(msg, category=RuntimeWarning)
            else:
                self._warn_if(PROD_ENV, 'enabled')

    def disable(self, *, permament: bool = False, warn: bool = True) -> None:
        """Disable all contracts.

        If `permament=True`, contracts are permanently disabled
        for the current interpreter runtime and cannot be turned on again.

        By default, deal will do a few sanity checks to make sure you haven't
        unintentionally disabled contracts on a test environment.
        Pass `warn=False` to disable this behavior.
        """
        if self.removed and permament:
            raise PERMAMENT_ERROR
        self.debug = False
        self.removed = permament
        if warn:
            self._warn_if(TEST_ENV, 'disabled')

    def _warn_if(self, env_vars: Mapping[str, str], state: str) -> None:
        for var, env in env_vars.items():
            if not os.environ.get(var):
                continue
            msg = f'It is {env} but deal is {state}. Is it intentional?'
            warnings.warn(msg, category=RuntimeWarning)
            return


state = _State()
reset = state.reset
enable = state.enable
disable = state.disable



from __future__ import annotations

import ast
from itertools import chain
from typing import TYPE_CHECKING

from .common import TOKENS, get_name, traverse


if TYPE_CHECKING:
    import astroid


def uses_result(validator) -> bool:
    """Returns True if the validator uses `result` argument.

    It may directly list `result` as the argument,
    or it may be a simple validator and use `_.result` in the body.
    """
    if not isinstance(validator, TOKENS.LAMBDA):
        return True
    if _has_result_arg(validator):
        return True
    if _is_simple_validator(validator):
        return _simple_uses_result(validator)
    return False


def _has_result_arg(validator: ast.Lambda | astroid.Lambda) -> bool:
    """Returns True if arguments list of the function has `result` argument.
    """
    if isinstance(validator, ast.Lambda):
        args = chain(
            validator.args.args,
            validator.args.kwonlyargs,
        )
        arg1: ast.arg
        for arg1 in args:
            if arg1.arg == 'result':
                return True
        return False
    else:
        args = chain(
            validator.args.args,
            validator.args.kwonlyargs,
        )
        arg2: astroid.AssignName
        for arg2 in args:
            if arg2.name == 'result':
                return True
        return False


def _is_simple_validator(validator: ast.Lambda | astroid.Lambda) -> bool:
    arg_names: list[str]
    if isinstance(validator, ast.Lambda):
        arg_names = [arg.arg for arg in validator.args.args]
    else:
        arg_names = [arg.name for arg in validator.args.args]
    return arg_names == ['_']


def _simple_uses_result(validator: ast.Lambda | astroid.Lambda) -> bool:
    """For simple validator, check if `_.result` is used in the body.
    """
    for node in traverse(body=[validator.body]):
        if get_name(node) == '_.result':
            return True
    return False



import ast
from contextlib import suppress
from pathlib import Path
from textwrap import dedent

import pytest

import deal
from deal._imports import DealLoader, deactivate, get_name


@pytest.mark.parametrize('text, expected', [
    ('name', 'name'),
    ('left.right', 'left.right'),

    ('left().right', None),
    ('1', None),
])
def test_get_name(text, expected):
    tree = ast.parse(text)
    print(ast.dump(tree))
    assert isinstance(tree, ast.Module)
    expr = tree.body[0].value
    assert get_name(expr=expr) == expected


def test_get_contracts():
    text = """
        import deal
        a = 1
        1 / 0
        not_a_deal.module_load(something)
        deal.module_load(deal.pure)
        """
    text = dedent(text)
    tree = ast.parse(text)
    print(ast.dump(tree))
    nodes = DealLoader._get_contracts(tree=tree)
    assert [get_name(node) for node in nodes] == ['deal.pure']


@pytest.mark.parametrize('text, expected', [
    ('deal.pure', deal.pure),
    # ('deal.has()', deal.has),
    ('deal.pre(something)', None),
    ('not_a_deal.pure', None),
    ('deal.typo', None),
])
def test_exec_contract(text, expected):
    tree = ast.parse(text)
    print(ast.dump(tree))
    assert isinstance(tree, ast.Module)
    actual = DealLoader._exec_contract(node=tree.body[0].value)
    assert actual == expected


class FakeError(Exception):
    pass


class FakeModule:
    pass


class SubLoader:
    def __init__(self, ok, text):
        self.ok = ok
        self.text = text

    def get_source(self, module):
        assert module == 'FakeModule'
        return self.text

    def exec_module(self, module):
        assert module is FakeModule
        if not self.ok:
            raise FakeError
        print(1)


def test_exec_module():
    text = """
        import deal
        deal.module_load(deal.has())
        print(1)
        """
    text = dedent(text)

    with pytest.raises(FakeError):
        DealLoader(loader=SubLoader(ok=False, text=text)).exec_module(FakeModule)

    with pytest.raises(deal.SilentContractError):
        DealLoader(loader=SubLoader(ok=True, text=text)).exec_module(FakeModule)


def test_not_path_loader():
    class SubLoaderNoGetSource:
        def exec_module(self, module):
            assert module is FakeModule
            raise FakeError

    with pytest.raises(FakeError):
        DealLoader(loader=SubLoaderNoGetSource()).exec_module(FakeModule)


def test_exec_module_invalid_contract():
    text = """
        import deal
        deal.module_load(deal.pre(something))
        print(1)
        """
    text = dedent(text)
    with pytest.raises(RuntimeError, match='unsupported contract:.+'):
        DealLoader(loader=SubLoader(ok=False, text=text)).exec_module(FakeModule)


def test_exec_module_invalid_contract_called():
    text = """
        import deal
        deal.module_load(something())
        print(1)
        """
    text = dedent(text)
    with pytest.raises(RuntimeError, match='unsupported contract:.+'):
        DealLoader(loader=SubLoader(ok=False, text=text)).exec_module(FakeModule)


def test_exec_module_no_contracts():
    text = """
        import deal
        print(1)
        """
    text = dedent(text)
    DealLoader(loader=SubLoader(ok=True, text=text)).exec_module(FakeModule)
    with pytest.raises(FakeError):
        DealLoader(loader=SubLoader(ok=False, text=text)).exec_module(FakeModule)


def test_exec_module_no_source():
    text = None
    DealLoader(loader=SubLoader(ok=True, text=text)).exec_module(FakeModule)
    with pytest.raises(FakeError):
        DealLoader(loader=SubLoader(ok=False, text=text)).exec_module(FakeModule)


def test_module_load():
    assert deal.activate()
    try:
        deal.module_load(deal.pure)
    finally:
        assert deactivate()

    with pytest.raises(RuntimeError):
        deal.module_load(deal.pure)


def test_module_load_no_contracts():
    with pytest.raises(RuntimeError):
        deal.module_load()


def test_activate():
    try:
        assert deal.activate()
        assert not deal.activate()
    finally:
        assert deactivate()
        assert not deactivate()


def test_smoke_pure():
    text = """
        import deal
        deal.module_load(deal.pure)
        print(1)
        """
    text = dedent(text)
    Path('tmp123.py').write_text(text)

    try:
        assert deal.activate()
        with pytest.raises(deal.SilentContractError):
            __import__('tmp123')
    finally:
        assert deactivate()
        assert not deactivate()
        with suppress(FileNotFoundError):
            Path('tmp123.py').unlink()


def test_smoke_has():
    text = """
        import urllib3
        import deal

        deal.module_load(deal.has('stdout'))
        # stdout is ok
        print(1)

        # network is not ok
        http = urllib3.PoolManager()
        http.request('GET', 'http://httpbin.org/robots.txt')
        """
    text = dedent(text)
    Path('tmp123.py').write_text(text)

    try:
        assert deal.activate()
        with pytest.raises(deal.OfflineContractError):
            __import__('tmp123')
    finally:
        assert deactivate()
        assert not deactivate()
        Path('tmp123.py').unlink()



from typing import List

import deal


# In short signature, `_` is a `dict` with access by attributes.
# Hence it has all dict attributes. So, if argument we need conflicts
# with a dict attribute, use getitem instead of getattr.
# In the example below, we use `_['items']` instead of `_.items`.

@deal.post(lambda result: result >= 0)
# if count is not zero, `item` appears in `items` at least once.
@deal.ensure(lambda _: _.result == 0 or _['item'] in _['items'])
# if count is zero, `item` is not in `items`
@deal.ensure(lambda _: _.result != 0 or _['item'] not in _['items'])
@deal.has()
def count(items: List[str], item: str) -> int:
    """How many times `item` appears in `items`
    """
    return items.count(item)


test_count = deal.cases(count)



from asyncio import iscoroutinefunction
from inspect import isgeneratorfunction

import deal

from .helpers import run_sync


def test_example_is_not_triggered_in_runtime():
    @deal.example(lambda: False)
    @deal.example(lambda: 1 / 0 == 0)
    def f1() -> bool:
        return True

    assert f1() is True


def test_example_does_not_break_iterator():
    @deal.example(lambda: False)
    @deal.example(lambda: 1 / 0 == 0)
    def f1():
        yield True

    assert isgeneratorfunction(f1) is True
    assert next(f1()) is True
    assert list(f1()) == [True]


def test_example_does_not_break_async():
    @deal.example(lambda: False)
    @deal.example(lambda: 1 / 0 == 0)
    async def f1() -> bool:
        return True

    assert iscoroutinefunction(f1) is True
    assert run_sync(f1()) is True



import pytest

import deal

from .helpers import run_sync


def test_parameters_and_result_fulfill_constact():
    @deal.ensure(lambda a, b, result: a > 0 and b > 0 and result != 'same number')
    def func(a, b) -> str:
        if a == b:
            return 'same number'
        else:
            return 'different numbers'

    assert func(1, 2) == 'different numbers'
    with pytest.raises(deal.PostContractError):
        func(0, 1)
    with pytest.raises(deal.PostContractError):
        func(1, 0)
    with pytest.raises(deal.PostContractError):
        func(1, 1)


def test_simple_signature():
    @deal.ensure(lambda _: _.a > 0 and _.b > 0 and _.result != 'same number')
    def func(a, b) -> str:
        if a == b:
            return 'same number'
        else:
            return 'different numbers'

    assert func(1, 2) == 'different numbers'
    with pytest.raises(deal.PostContractError):
        func(0, 1)
    with pytest.raises(deal.PostContractError):
        func(1, 0)
    with pytest.raises(deal.PostContractError):
        func(1, 1)


def test_decorating_async_function():
    @deal.ensure(lambda a, b, result: a > 0 and b > 0 and result != 'same number')
    async def func(a, b) -> str:
        if a == b:
            return 'same number'
        else:
            return 'different numbers'

    assert run_sync(func(1, 2)) == 'different numbers'
    with pytest.raises(deal.PostContractError):
        run_sync(func(0, 1))


def test_decorating_generator():
    @deal.ensure(lambda x, y, result: result > y ** 2)
    def multiply(x, y):
        yield x * y
        yield x * y * 2
        yield x * y * 4

    assert list(multiply(2, 1)) == [2, 4, 8]
    with pytest.raises(deal.PostContractError):
        list(multiply(2, 2))


def test_recursive_contracts_ok():
    @deal.ensure(lambda a, b, result: add(result, b) == a)
    def sub(a, b):
        return a - b

    @deal.ensure(lambda a, b, result: sub(result, b) == a)
    def add(a, b):
        return a + b

    assert sub(5, 3) == 2
    assert add(2, 3) == 5



from __future__ import annotations

import ast
import re
import tokenize
from pathlib import Path
from typing import Iterator

from .. import __version__
from ._error import Error
from ._func import Func
from ._rules import FuncRule, ModuleRule, rules
from ._stub import StubsManager


try:
    import astroid
except ImportError:
    astroid = None


# based on regex from the source code of flake8
RE_NOQA = re.compile(
    r'# noqa(?::[\s]?(?P<codes>([A-Z]+[0-9]*(?:[,\s]+)?)+))?',
    re.IGNORECASE,
)
RE_DIGITS = re.compile(r'[0-9]+')
PREFIXES = ('DEAL', 'DEA1', 'DEL', 'DEA', 'DIL')


class Checker:
    """The entry point for the linter, compatible with flake8 plugins interface.
    """
    __slots__ = ('_tree', '_filename', '_tokens', '_stubs')
    name = 'deal'
    version = __version__
    _rules = rules

    def __init__(
        self,
        tree: ast.Module,
        file_tokens: list[tokenize.TokenInfo] | None = None,
        filename: str = 'stdin',
    ) -> None:
        self._tree = tree
        self._filename = filename
        self._tokens = file_tokens

        paths = list(StubsManager.default_paths)
        if filename != 'stdin':
            paths.append(Path(filename).absolute().parent)
        self._stubs = StubsManager(paths=paths)

    @classmethod
    def from_path(cls, path: Path) -> Checker:
        source = path.read_text()
        with path.open('rb') as stream:
            tokens = list(tokenize.tokenize(stream.readline))
        return cls(
            tree=ast.parse(source),
            file_tokens=tokens,
            filename=str(path),
        )

    def run(self) -> Iterator[tuple]:
        for error in self.get_errors():
            yield tuple(error) + (type(self),)

    def get_funcs(self) -> list[Func]:
        if self._filename == 'stdin' or astroid is None:
            return Func.from_ast(tree=self._tree)
        text = Path(self._filename).read_text()
        try:
            tree = astroid.parse(text)
        except astroid.AstroidSyntaxError:
            return Func.from_ast(tree=self._tree)
        return Func.from_astroid(tree)

    def get_errors(self) -> Iterator[Error]:
        reported = set()
        for func in self.get_funcs():
            for rule in self._rules:
                if not isinstance(rule, FuncRule):
                    continue
                for error in rule(func=func, stubs=self._stubs):
                    hs = hash(error)
                    if hs in reported:
                        continue
                    noqa = self._get_noqa(error.row)
                    if f'{error.code:03d}'.startswith(noqa):
                        continue
                    reported.add(hs)
                    yield error

        for rule in self._rules:
            if not isinstance(rule, ModuleRule):
                continue
            yield from rule(tree=self._tree)

    def _get_noqa(self, line_number: int) -> tuple[str, ...]:
        """Extract codes defined in `noqa` comments (with DEAL prefix removed).
        """
        comment = self._get_comment(line_number)
        if not comment:
            return ()
        match = RE_NOQA.search(comment)
        if match is None:
            return ()
        raw_codes = (match.group('codes') or '').split(',')
        clean_codes = []
        for code in raw_codes:
            code = code.strip().upper()
            if not code.startswith(PREFIXES):
                print(repr(code), match)
                continue
            match = RE_DIGITS.search(code)
            clean_codes.append(match.group(0) if match else '')
        return tuple(clean_codes)

    def _get_comment(self, line_number: int) -> str | None:
        if self._tokens is None:
            return None
        for token in self._tokens:
            if token.start[0] != line_number:
                continue
            if token.type != tokenize.COMMENT:
                continue
            return token.string
        return None



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any


if TYPE_CHECKING:
    import vaa


class Scheme(ABC):
    """A base class for implementing a custom validator.

    The custom validator should implement `is_valid` method.
    The method should return True if the data is valid.
    Otherwise, it should set `errors` attribute and return False.
    """
    __slots__ = ('data', 'errors')
    data: dict[str, Any]
    errors: list[vaa.Error]

    def __init__(self, data: dict[str, Any]) -> None:
        self.data = data
        self.errors = []

    @abstractmethod
    def is_valid(self) -> bool:
        raise NotImplementedError



from __future__ import annotations

import ast
from textwrap import dedent
from typing import TypeVar

from deal.linter._func import Func


T = TypeVar('T')


def first(funcs: list[T]) -> T:
    assert len(funcs) == 1
    return funcs[0]


def funcs_from_ast(text: str) -> list[Func]:
    text = dedent(text).strip()
    tree = ast.parse(text)
    print(ast.dump(tree))
    return Func.from_ast(tree)


def funcs_from_astroid(text: str) -> list[Func]:
    import astroid

    text = dedent(text).strip()
    tree = astroid.parse(text)
    print(tree.repr_tree())
    return Func.from_astroid(tree)



import ast

import pytest

from deal.linter._extractors import get_imports


try:
    import astroid
except ImportError:
    astroid = None


@pytest.mark.parametrize('text, expected', [
    ('from deal import pre', ('deal', )),
])
def test_get_imports_simple(text, expected):
    if astroid is not None:
        tree = astroid.parse(text)
        print(tree.repr_tree())
        returns = tuple(r.value for r in get_imports(body=tree.body))
        assert returns == expected

    ast_tree = ast.parse(text)
    print(ast.dump(ast_tree))
    returns = tuple(r.value for r in get_imports(body=ast_tree.body))
    assert returns == expected



import ast

import pytest

from deal.linter._extractors import get_example
from deal.linter._extractors.value import UNKNOWN


try:
    import astroid
except ImportError:
    astroid = None


@pytest.mark.parametrize('text, expected', [
    ('f(1) == 2', ([1], {}, 2)),
    ('f(1, 2) == 3', ([1, 2], {}, 3)),
    ('f(one=1) == 2', ([], {'one': 1}, 2)),
    ('f(one=1, two=2) == 3', ([], {'one': 1, 'two': 2}, 3)),
    ('f(3, 4, one=1, two=2) == 5', ([3, 4], {'one': 1, 'two': 2}, 5)),
    ('f() is None', ([], {}, None)),

    # cannot extract function result
    ('f() == abc', ([], {}, UNKNOWN)),
    ('f() < 4', ([], {}, UNKNOWN)),
    ('f() < 4 < 5', ([], {}, UNKNOWN)),

    # cannot extract example
    ('f()', None),
    ('fu() == 1', None),
    ('f(a) == 1', None),
    ('f(a=b) == 1', None),
    ('f(1)', None),
    ('f() + 4', None),
    ('hello', None),
    ('a == b', None),
    ('a == 4', None),
    ('4 == a', None),
])
def test_get_asserts_simple(text, expected):
    if astroid is not None:
        node = astroid.extract_node(text)
        print(node.repr_tree())
        result = get_example(node, func_name='f')
        assert result == expected

    ast_node = ast.parse(text).body[0]
    assert isinstance(ast_node, ast.Expr)
    print(ast.dump(ast_node))
    result = get_example(ast_node.value, func_name='f')
    assert result == expected



import random
from typing import List

import deal


# the list cannot be empty
@deal.pre(lambda items: bool(items))
# result is an element withit the given list
@deal.ensure(lambda items, result: result in items)
@deal.has('random')
def choice(items: List[str]) -> str:
    """Get a random element from the given list.
    """
    return random.choice(items)


test_choice = deal.cases(choice)



import ast
import sys
from textwrap import dedent

import pytest

from deal.linter._func import Func

from .helpers import first, funcs_from_astroid


try:
    import astroid
except ImportError:
    astroid = None


TEXT = """
    import deal

    @deal.post(lambda x: x > 0)
    @deal.raises(ValueError, UnknownError, idk_what_is_it())
    @notadeal.raises(KeyError)
    @deal.notaraises(KeyError)
    @something
    @another.deco
    @something()
    def f(x):
        return x

    def h(x):
        return x

    @something()
    def k(x):
        return x
"""


def test_from_text():
    funcs = Func.from_text(dedent(TEXT))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_from_ast():
    funcs = Func.from_ast(ast.parse(dedent(TEXT)))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_from_astroid():
    funcs = Func.from_astroid(astroid.parse(dedent(TEXT)))
    assert len(funcs) == 3
    assert len(funcs[0].contracts) == 2


def test_from_ast_methods():
    text = dedent("""
        class A:
            p = 1
            @deal.post(lambda x: x > 0)
            def f(self, x):
                return x
    """)
    funcs = Func.from_ast(ast.parse(text))
    assert len(funcs) == 1
    assert len(funcs[0].contracts) == 1


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_from_astroid_methods():
    text = dedent("""
        class A:
            p = 1
            @deal.post(lambda x: x > 0)
            def f(self, x):
                return x
    """)
    funcs = Func.from_astroid(astroid.parse(text))
    assert len(funcs) == 1
    assert len(funcs[0].contracts) == 1


def test_repr():
    funcs1 = Func.from_ast(ast.parse(dedent(TEXT)))
    assert repr(funcs1[0]) == 'Func(post, raises)'

    if astroid is not None:
        funcs2 = Func.from_astroid(astroid.parse(dedent(TEXT)))
        assert repr(funcs2[0]) == 'Func(post, raises)'


mark38 = pytest.mark.skipif(sys.version_info < (3, 8), reason='old python')


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
@pytest.mark.parametrize('signature, expected', [
    ('self', True),
    ('self, a', True),
    ('self, *a, **k', True),
    ('self, *, a', True),

    ('a', False),
    ('a, b', False),
    ('*self', False),
    ('**self', False),

    pytest.param('self, /, a', True, marks=mark38),
    pytest.param('self, a, /, b', True, marks=mark38),
    pytest.param('a, /, b', False, marks=mark38),
])
def test_has_self(signature, expected):
    text = f"""
        @deal.post(lambda x: x > 0)
        def f({signature}):
            return x
    """
    func = first(funcs_from_astroid(text))
    assert func.has_self is expected



from typing import List

import deal


# if you have more than 2-3 contracts,
# consider moving them from decorators into separate variable
# like this:
contract_for_index_of = deal.chain(
    # result is an index of items
    deal.post(lambda result: result >= 0),
    deal.ensure(lambda items, item, result: result < len(items)),
    # element at this position matches item
    deal.ensure(
        lambda items, item, result: items[result] == item,
        message='invalid match',
    ),
    # element at this position is the first match
    deal.ensure(
        lambda items, item, result: not any(el == item for el in items[:result]),
        message='not the first match',
    ),
    # LookupError will be raised if no elements found
    deal.raises(LookupError),
    deal.reason(LookupError, lambda items, item: item not in items),
    # no side-effects
    deal.has(),
)


@contract_for_index_of
def index_of(items: List[int], item: int) -> int:
    for index, el in enumerate(items):
        if el == item:
            return index
    raise LookupError


test_index_of = deal.cases(index_of)



from __future__ import annotations

import ast
import builtins
import enum
from copy import copy
from pathlib import Path
from types import CodeType
from typing import Iterable, Union

from .._cached_property import cached_property


try:
    import astroid
except ImportError:
    pass


TEMPLATE = (Path(__file__).parent / '_template.py').read_text()
CONTRACT_INDEX = 3
FUNC_INDEX = 4


class Category(enum.Enum):
    ENSURE = 'ensure'
    EXAMPLE = 'example'
    HAS = 'has'
    POST = 'post'
    PRE = 'pre'
    PURE = 'pure'
    RAISES = 'raises'
    SAFE = 'safe'
    INHERIT = 'inherit'

    @property
    def brackets_optional(self) -> bool:
        return self in {Category.SAFE, Category.PURE}


class NoValidatorError(Exception):
    pass


class Contract:
    args: tuple[ast.expr | astroid.NodeNG, ...]
    kwargs: tuple[ast.keyword | astroid.Keyword, ...]
    category: Category
    func_args: ast.arguments
    context: dict[str, ast.stmt]
    line: int

    def __init__(
        self,
        args: Iterable[ast.expr | astroid.NodeNG],
        kwargs: Iterable[ast.keyword | astroid.Keyword],
        category: Category,
        func_args: ast.arguments,
        context: dict[str, ast.stmt] | None = None,
        line: int = 0,
    ) -> None:
        self.args = tuple(args)
        self.kwargs = tuple(kwargs)
        self.category = category
        self.func_args = func_args
        self.context = context or dict()
        self.line = line

    @cached_property
    def validator(self) -> ast.AST:
        """The validator converted into ast.
        """
        contract = self.raw_validator
        if isinstance(contract, ast.AST):
            return contract
        # convert astroid node to ast node
        contract = self._resolve_name(contract)
        return ast.parse(contract.as_string()).body[0]

    @cached_property
    def raw_validator(self) -> ast.expr | astroid.NodeNG:
        if self.args:
            return self.args[0]
        for kwarg in self.kwargs:
            if kwarg.arg == 'validator':
                return kwarg.value
        raise NoValidatorError('cannot find validator for contract')

    @cached_property
    def arguments(self) -> frozenset[str]:
        """Contract function arguments names.

        Useful for resolving external dependencies.
        """
        func = self.validator
        if isinstance(func, ast.Expr):
            func = func.value
        if not isinstance(func, (ast.FunctionDef, ast.Lambda)):
            return frozenset()
        args = func.args
        result = set()
        for arg in args.args:
            result.add(arg.arg)
        for arg in args.kwonlyargs:
            result.add(arg.arg)
        if args.vararg:
            result.add(args.vararg.arg)
        if args.kwarg:
            result.add(args.kwarg.arg)
        return frozenset(result)

    @cached_property
    def dependencies(self) -> frozenset[str]:
        """Names that are defined outside of the contract body.

        1. Excludes built-in objects.
        1. Excludes contract function arguments.
        """
        deps = set()
        for node in ast.walk(self.validator):
            if not isinstance(node, ast.Name):
                continue
            if hasattr(builtins, node.id):
                continue
            if node.id in self.arguments:
                continue
            deps.add(node.id)
        return frozenset(deps)

    @staticmethod
    def _resolve_name(contract: astroid.NodeNG) -> astroid.NodeNG:
        if not isinstance(contract, astroid.Name):
            return contract
        definitions = contract.lookup(contract.name)[1]
        if not definitions:
            return contract
        definition = definitions[0]
        if isinstance(definition, astroid.FunctionDef):
            return definition
        if isinstance(definition, astroid.AssignName):
            return definition.parent.value
        # resolved into something tricky, live with it
        return contract  # pragma: no cover

    @cached_property
    def exceptions(self) -> list[Union[str, type[BaseException]]]:
        from ._extractors import get_name

        excs = []
        for expr in self.args:
            name = get_name(expr)
            if not name:
                continue
            exc = getattr(builtins, name, name)
            excs.append(exc)
        return excs

    @cached_property
    def module(self) -> ast.Module:
        module = ast.parse(TEMPLATE)

        # inject function signature
        func = ast.Lambda(
            args=self.func_args,
            body=ast.Set(elts=[]),
        )
        module.body[FUNC_INDEX].value = func  # type: ignore

        # collect definitions for contract external deps
        deps: list[ast.stmt] = []
        for dep in self.dependencies:
            definition = self.context.get(dep)
            if not definition:
                continue
            deps.append(definition)

        # inject contract if contract is a function
        contract = copy(self.validator)
        if isinstance(contract, ast.FunctionDef):
            contract.body = deps + contract.body
            # if contract is function, add it's definition and assign it's name
            # to `contract` variable.
            module.body = [contract] + module.body
            module.body[FUNC_INDEX].value = ast.Name(  # type: ignore[attr-defined]
                id=contract.name,
                ctx=ast.Load(),
            )
            return module

        if isinstance(contract, ast.Expr):
            contract = contract.value

        # Inject contract if contract is a lambda.
        # We have to rebuild lambda into a function
        # to inject dependencies inside the body.
        if isinstance(contract, ast.Lambda):
            body = list(deps)
            return_node = ast.Return(
                value=contract.body,
            )
            body.append(return_node)
            module.body[CONTRACT_INDEX] = ast.FunctionDef(
                name='contract',
                args=contract.args,
                body=body,
                decorator_list=[],
            )
            return module

        # inject contract if contract is an unknown expression
        module.body[CONTRACT_INDEX].value = contract  # type: ignore[attr-defined]
        return module

    @cached_property
    def bytecode(self) -> CodeType:
        module = ast.fix_missing_locations(self.module)
        return compile(module, filename='<ast>', mode='exec')

    def run(self, *args, **kwargs):
        globals = dict(args=args, kwargs=kwargs)
        exec(self.bytecode, globals)
        return globals['result']

    def __repr__(self) -> str:
        return f'{type(self).__name__}({self.category.value})'



import pytest

import deal

from .helpers import run_sync


def test_safe():
    func = deal.safe(lambda x: 1 / x)
    func(2)
    with pytest.raises(deal.RaisesContractError):
        func(0)


def test_decorating_async_function():
    @deal.safe
    async def func(x):
        return 10 / x

    assert run_sync(func(2)) == 5
    with pytest.raises(deal.RaisesContractError):
        run_sync(func(0))



from pathlib import Path

import pytest

from deal._cli._common import get_paths


def test_get_paths(tmp_path: Path):
    (tmp_path / 'subdir').mkdir()
    (tmp_path / 'subdir' / '__pycache__').mkdir()
    (tmp_path / '.hidden').mkdir()

    (tmp_path / 'setup.py').touch()
    (tmp_path / 'subdir' / 'ex.py').touch()
    (tmp_path / '.hidden' / 'ex.py').touch()
    (tmp_path / 'subdir' / '__pycache__' / 'ex.py').touch()
    (tmp_path / 'setup.pl').touch()
    actual = {p.relative_to(tmp_path) for p in get_paths(tmp_path)}
    assert actual == {Path('setup.py'), Path('subdir/ex.py')}

    with pytest.raises(FileNotFoundError):
        list(get_paths(tmp_path / 'not_exists'))



import ast
from textwrap import dedent

import pytest

from deal.linter._extractors import get_markers
from deal.linter._stub import StubsManager


try:
    import astroid
except ImportError:
    astroid = None


@pytest.mark.parametrize('text, expected', [
    ('print(1)', ('stdout', )),
    ('print(1, sep=sys.stderr)', ('stdout', )),
    ('print(1, file=sys.stdout)', ('stdout', )),
    ('print(1, file=sys.stderr)', ('stderr', )),
    ('print(1, file=stream)', ()),

    ('import sys\nsys.stdout.write(1)', ('stdout', )),
    ('import sys\nsys.stderr.write(1)', ('stderr', )),

    ('open("fpath", "w")', ('write', )),
    ('open("fpath", mode="w")', ('write', )),
    ('with open("fpath", "w") as f: ...', ('write', )),

    ('with something: ...', ()),     # uninferrable `with` test
    ('something().anything()', ()),  # complex call, cannot get name

    ('with open("fpath") as f: ...', ('read', )),
    ('with open("fpath", "r") as f: ...', ('read', )),
    ('with open("fpath", mode="r") as f: ...', ('read', )),
    ('open("fpath", "r")', ('read', )),
    ('open("fpath")', ('read', )),
    ('open("fpath", encoding="utf8")', ('read', )),

    ('input()', ('stdin', )),
    ('input("say hi: ")', ('stdin', )),
    ('sys.stdin.read()', ('stdin', )),
    ('sys.stdin.read(10)', ('stdin', )),

    ('random.randint(10)', ('random', )),
    ('random.randrange(10)', ('random', )),
    ('random.random(10)', ('random', )),
    ('randrange(10)', ('random', )),

    ('os.system("echo")', ('syscall', )),
    ('os.execvp("echo", ["hi"])', ('syscall', )),
    ('subprocess.run("echo")', ('syscall', )),
    ('subprocess.call("echo")', ('syscall', )),
    ('subprocess.call(["echo"])', ('syscall', )),
    ('proc = subprocess.Popen("echo")', ('syscall', )),
    ('subprocess.SubprocessError()', ()),

    ('time.time()', ('time', )),
    ('time()', ('time', )),
    ('time_ns()', ('time', )),
    ('os.times()', ('time', )),
    ('datetime.now()', ('time', )),
    ('now()', ()),
])
def test_io_hardcoded(text, expected):
    if astroid is not None:
        tree = astroid.parse(text)
        print(tree.repr_tree())
        tokens = list(get_markers(body=tree.body))
        markers = tuple(t.marker for t in tokens if t.marker != 'import')
        assert markers == expected

    ast_tree = ast.parse(text)
    print(ast.dump(ast_tree))
    tokens = list(get_markers(body=ast_tree.body))
    markers = tuple(t.marker for t in tokens if t.marker != 'import')
    assert markers == expected


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
@pytest.mark.parametrize('text, expected', [
    ('from pathlib import Path\np = Path()\np.write_text("lol")', ('write', )),
    ('from pathlib import Path\np = Path()\np.open("w")', ('write', )),
    ('from pathlib import Path\np = Path()\nwith p.open("w"): ...', ('write', )),

    ('from pathlib import Path\np = Path()\np.open("r")', ()),          # allowed mode
    ('from pathlib import Path\np = Path()\np.open(mode="r")', ()),     # allowed mode
    ('from pathlib import Path\np = Path()\np.read_text()', ()),    # read, not write
    ('from pathlib import Path\np = Path()\np.open()', ()),         # implicit read

    # not pathlib
    ('with something.open("w"): ...', ()),
    ('something = file\nwith something.open("w"): ...', ()),
    ('class Path:\n def write_text(): pass\np = Path()\np.write_text()', ()),
    ('class Path:\n def write_text(): pass\nPath.write_text()', ()),
])
def test_io_infer(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens if t.marker != 'import')
    assert markers == expected


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
@pytest.mark.parametrize('text, expected', [
    ('from random import choice \nchoice([1,2])', ('random', )),
    ('choice([1,2])', ()),
    ('choice = lambda:0 \nchoice()', ()),
    ('class A:\n def b(self): pass \nchoice = A().b \nchoice()', ()),
])
def test_other_infer(text, expected):
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens if t.marker != 'import')
    assert markers == expected


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_socket():
    text = """
        import socket
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((HOST, PORT))
    """
    tree = astroid.parse(text)
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('import', 'network')


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_asyncio_socket():
    text = """
        import asyncio
        r, w = await asyncio.open_connection('127.0.0.1', 8888)
    """
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('import', 'network')


@pytest.mark.parametrize('text, expected', [
    ('global a', ('global', )),
    ('global a, b, c', ('global', )),

    ('nonlocal a', ('global', )),
    ('nonlocal a, b, c', ('global', )),

    ('import a', ('import', )),
    ('import a as b', ('import', )),
    ('import a as b, c', ('import', )),

    ('from a import b', ('import', )),
    ('from a import b as c', ('import', )),

    ('__import__("something")', ('import', )),
])
def test_get_globals_simple(text, expected):
    if astroid is not None:
        tree = astroid.parse(text)
        print(tree.repr_tree())
        tokens = list(get_markers(body=tree.body))
        markers = tuple(t.marker for t in tokens)
        assert markers == expected

    ast_tree = ast.parse(text)
    print(ast.dump(ast_tree))
    tokens = list(get_markers(body=ast_tree.body))
    markers = tuple(t.marker for t in tokens)
    assert markers == expected


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_io_recursive_analize_body():
    text = """
    def inner(text):
        print(text)

    def outer():
        inner('hello')
    """
    text = dedent(text)
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body[-1].body))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('stdout', )


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_io_recursive_explicit_markers():
    text = """
    @deal.has('io', invalid)
    @deal.raises()
    def inner(text):
        noting()

    def outer():
        inner('hello')
    """
    text = dedent(text)
    tree = astroid.parse(text)
    print(tree.repr_tree())
    tokens = list(get_markers(body=tree.body[-1].body))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('io', )


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_markers_from_stubs():
    text = """
    import ast

    nothing = None

    def do_nothing():
        pass

    @deal.has()
    def inner(text):
        ast.walk(None)
        not_resolvable()
        do_nothing()
        nothing()
    """
    text = dedent(text)
    tree = astroid.parse(text)
    print(tree.repr_tree())
    stubs = StubsManager()
    tokens = list(get_markers(body=tree.body[-1].body, stubs=stubs))
    markers = tuple(t.marker for t in tokens)
    assert markers == ('import', )



import ast
import subprocess
import sys
from pathlib import Path
from textwrap import dedent

import pytest

from deal.linter import Checker


try:
    import astroid
except ImportError:
    astroid = None

try:
    import flake8
except ImportError:
    flake8 = None


TEXT = """
import deal


@deal.post(lambda x: x > 0)
def test():
    return -1


@deal.raises(ValueError)
def test2():
    1 / 0
    raise ValueError
    raise KeyError
""".strip()

EXPECTED = [
    (6, 11, 'DEL012 post contract error (-1)', Checker),
    (11, 8, 'DEL021 raises contract error (ZeroDivisionError)', Checker),
    (13, 10, 'DEL021 raises contract error (KeyError)', Checker),
]


def test_stdin():
    checker = Checker(tree=ast.parse(TEXT))
    errors = list(checker.run())
    assert errors == EXPECTED


def test_astroid_path(tmp_path: Path):
    path = tmp_path / 'test.py'
    path.write_text(TEXT)
    checker = Checker.from_path(path)
    errors = list(checker.run())
    assert errors == EXPECTED


def test_get_funcs_invalid_syntax(tmp_path: Path):
    """
    Atom IDE flake8 plugin can call flake8 with AST with correct syntax but with path
    to code with invalid syntax. In that case, we should ignore the file and fallback
    to the passed AST.
    """
    path = tmp_path / 'test.py'
    path.write_text('1/')
    checker = Checker(tree=ast.parse(TEXT), filename=str(path))
    errors = list(checker.run())
    assert errors == EXPECTED


def test_version():
    version = Checker(tree=None, filename='stdin').version  # type: ignore[arg-type]
    assert version.count('.') == 2
    assert not set(version) - set('0123456789.')


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_remove_duplicates(tmp_path: Path):
    text = """
        import deal

        def inner():
            raise TypeError
            raise TypeError

        @deal.raises()
        def outer():
            return inner()
    """
    path = tmp_path / 'test.py'
    path.write_text(dedent(text))
    checker = Checker.from_path(path)
    errors = list(checker.run())
    assert len(errors) == 1


@pytest.mark.skipif(sys.version_info < (3, 8, 1), reason='flake8 works only on 3.8.1+')
@pytest.mark.skipif(flake8 is None, reason='flake8 is not installed')
def test_flake8_integration(tmp_path: Path):
    path = tmp_path / 'test.py'
    path.write_text(TEXT + '\n')
    cmd = [sys.executable, '-m', 'flake8', str(tmp_path)]
    result = subprocess.run(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
    assert result.returncode == 1
    lines = result.stdout.decode().splitlines()
    assert len(EXPECTED) == len(lines), result.stderr
    for (lineno, col, error, _), line in zip(EXPECTED, lines):
        assert error in line
        assert f':{lineno}:{col + 1}' in line


@pytest.mark.parametrize('comment, should_fail', [
    ('', True),

    # different prefix sizes
    ('# noqa: DEAL021', False),
    ('# noqa: DEAL02', False),
    ('# noqa: DEAL0', False),
    ('# noqa: DEAL', False),

    # case insensitive
    ('# NoQA: DEAL', False),
    ('# NOQA: DEAL', False),
    ('# noqa: deal', False),

    # aliases
    ('# noqa: DEA021', False),
    ('# noqa: DEL021', False),
    ('# noqa: DIL021', False),
    ('# noqa: DEA0', False),
    ('# noqa: DEA', False),

    # mixed with other codes and garbage
    ('# noqa: W13, DEAL021', False),
    ('# noqa: DEAL021, W13', False),
    ('# noqa: DEAL021 # nosec # type: ignore', False),
    ('# nosec # type: ignore # noqa: DEAL021', False),

    # not this noqa
    ('# noqa: W13', True),
    ('# noqa: DEAL022', True),
    ('# noqa: D022', True),
    ('# oh hi mark', True),
    ('# type: ignore', True),
    ('# nosec', True),
])
def test_noqa(tmp_path: Path, comment, should_fail):
    text = f"""
        import deal
        @deal.safe
        def f():
            1/0 {comment}
    """
    path = tmp_path / 'test.py'
    path.write_text(dedent(text))
    checker = Checker.from_path(path)
    errors = list(checker.run())
    assert len(errors) == int(should_fail)



import re
from typing import NoReturn, TypeVar

import pytest

import deal


try:
    import hypothesis
except ImportError:
    hypothesis = None
else:
    import hypothesis.errors
    import hypothesis.strategies

try:
    import typeguard
except ImportError:
    typeguard = None


@deal.raises(ZeroDivisionError)
def div1(a: int, b: int) -> float:
    assert isinstance(a, int)
    assert isinstance(b, int)
    return a / b


@deal.pre(lambda a, b: a > 0)
@deal.pre(lambda a, b: b > 0, message='b must be positive')
def div2(a: int, b: int) -> float:
    assert isinstance(a, int)
    assert isinstance(b, int)
    assert a > 0
    assert b > 0
    return a / b


if hypothesis is not None:
    test_div1_short = deal.cases(div1)
    test_div2_short = deal.cases(div2)


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_short_version_is_discoverable(monkeypatch):
    from _pytest.python import PyCollector

    collector = PyCollector.__new__(PyCollector)
    monkeypatch.setattr(collector, '_matches_prefix_or_glob_option', lambda *args: True)
    test = deal.cases(div1)
    assert collector.istestfunction(test, 'test_div') is True


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_count():
    for count in (1, 10, 20, 50):
        cases = deal.cases(div1, count=count)
        assert len(list(cases)) == count

        cases = deal.cases(div2, count=count)
        assert len(list(cases)) == count


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_params_detected():
    for case in deal.cases(div1, count=10):
        assert set(case.kwargs) == {'a', 'b'}

    for case in deal.cases(div2, count=10):
        assert set(case.kwargs) == {'a', 'b'}


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_params_type():
    for case in deal.cases(div1, count=10):
        assert type(case.kwargs['a']) is int
        assert type(case.kwargs['b']) is int

    for case in deal.cases(div2, count=10):
        assert type(case.kwargs['a']) is int
        assert type(case.kwargs['b']) is int


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_params_ok_with_excs():
    results = []
    for case in deal.cases(div1, count=20):
        result = case()
        results.append(result)
    assert len(results) == 20
    assert any(r is NoReturn for r in results), 'no exception occured'


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_no_bad_examples():
    for case in deal.cases(div2, count=20):
        assert case.kwargs['a'] > 0
        assert case.kwargs['b'] > 0


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_return_type_checks():
    def div(a: int, b: int) -> int:
        return 1

    for case in deal.cases(div, count=20):
        case()

    def div(a: int, b: int) -> str:  # type: ignore[no-redef]
        return 1  # type: ignore[return-value]

    msg = re.escape('the return value (int) is not an instance of str')
    with pytest.raises(typeguard.TypeCheckError, match=msg):
        case = next(iter(deal.cases(div, count=20)))
        case()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_explicit_kwargs():
    def div(a: int, b: int):
        assert b == 4

    for case in deal.cases(div, kwargs=dict(b=4), count=20):
        case()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_explicit_strategy():
    def div(a: int, b: int):
        assert 0 <= b <= 4

    cases = deal.cases(
        div,
        kwargs=dict(b=hypothesis.strategies.integers(min_value=0, max_value=4)),
        count=20,
    )
    for case in cases:
        case()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_disable_type_checks():
    def bad(a: int) -> str:
        return a  # type: ignore[return-value]

    # type is wrong and checked
    cases = deal.cases(bad, count=1)
    case = next(iter(cases))
    msg = re.escape('the return value (int) is not an instance of str')
    with pytest.raises(typeguard.TypeCheckError, match=msg):
        case()

    # type is wrong and ignored
    cases = deal.cases(bad, count=1, check_types=False)
    case = next(iter(cases))
    case()

    def good(a: int) -> int:
        return a

    # type is good
    cases = deal.cases(good, count=1)
    case = next(iter(cases))
    case()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
@pytest.mark.skipif(typeguard is None, reason='typeguard is not installed')
def test_return_type():
    def identity(a) -> int:
        return a

    kwargs: dict = dict(kwargs={}, func=identity, exceptions=(), check_types=True)
    case = deal.TestCase(args=(12, ), **kwargs)
    case()

    case = deal.TestCase(args=('hi', ), **kwargs)
    msg = re.escape('the return value (str) is not an instance of int')
    with pytest.raises(typeguard.TypeCheckError, match=msg):
        case()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
@pytest.mark.skipif(typeguard is None, reason='typeguard is not installed')
def test_typecheck_explicit_kwargs():
    def identity(a: int) -> str:
        return 'ok'

    kwargs: dict = dict(args=(), func=identity, exceptions=(), check_types=True)
    case = deal.TestCase(kwargs={'a': 13}, **kwargs)
    case()

    case = deal.TestCase(kwargs={'a': 'hi'}, **kwargs)
    msg = re.escape('argument "a" (str) is not an instance of int')
    with pytest.raises(typeguard.TypeCheckError, match=msg):
        case()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_type_var():
    T = TypeVar('T')  # noqa: N806

    def identity(a: T, b) -> T:
        return b

    kwargs: dict = dict(kwargs={}, func=identity, exceptions=(), check_types=True)
    case = deal.TestCase(args=('ab', 'cd'), **kwargs)
    case()

    case = deal.TestCase(args=(['ab'], ['cd', 'ef']), **kwargs)
    case()


if hypothesis is not None:
    @deal.cases(div1)
    def test_decorator_div1_smoke(case):
        case()

    @deal.cases(div2)
    def test_decorator_div2_smoke(case):
        case()

    @deal.cases(div2)
    def test_decorator_rejects_bad(case):
        assert case.kwargs['a'] > 0
        assert case.kwargs['b'] > 0
        case()

    @deal.cases(div1, kwargs=dict(b=0))
    def test_decorator_suppress_raises(case):
        result = case()
        assert result is NoReturn


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_repr():
    def fn():
        pass

    cases = deal.cases(fn)
    assert repr(cases) == 'deal.cases(fn, count=50)'

    cases = deal.cases(fn, count=13, seed=2, kwargs=dict(a=2))
    assert repr(cases) == "deal.cases(fn, count=13, seed=2, kwargs={'a': 2})"


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_seed():
    c1 = list(deal.cases(div1, seed=12, count=20))
    c2 = list(deal.cases(div1, seed=12, count=20))
    assert c1 == c2

    c3 = list(deal.cases(div1, seed=34, count=20))
    assert c2 != c3


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_run_ok():
    test = deal.cases(div1)
    res = test()
    assert res is None


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_run_fail():
    @deal.safe
    def div(a: int, b: int):
        return a / b

    test = deal.cases(div, seed=1)
    with pytest.raises(deal.RaisesContractError):
        test()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_fuzz_propagate():
    @deal.safe
    def div(a: str):
        assert type(a) is str
        raise ZeroDivisionError

    cases = deal.cases(div, seed=1)
    with pytest.raises(deal.RaisesContractError):
        cases(b'g`\x00\x00\x00\x00\x00\x00\x00\x00')


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_fuzz_bad_input():
    @deal.safe
    def div(a: str):
        raise ZeroDivisionError

    cases = deal.cases(div, seed=1)
    res = cases(b'')
    assert res is None


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_pass_fixtures():
    @deal.safe
    def div(a: str):
        raise ZeroDivisionError

    @deal.cases(div)
    def test_div1(case, fixt):
        assert fixt == 13
        case()

    with pytest.raises(deal.RaisesContractError):
        test_div1(fixt=13)

    with pytest.raises(deal.RaisesContractError):
        test_div1(13)


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_reproduce_failure():
    @deal.safe
    def div(a: str):
        assert a == ''
        raise ZeroDivisionError

    @hypothesis.reproduce_failure(hypothesis.__version__, b'AAA=')
    @deal.cases(div)
    def test_div(case):
        assert case.kwargs == dict(a='')
        case()

    with pytest.raises(deal.RaisesContractError):
        test_div()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_example():
    @deal.example(lambda: double1(2) == 4)
    def double1(x: int) -> int:
        return x * 2

    for case in deal.cases(double1):
        case()

    @deal.example(lambda: double2(2) == 5)
    def double2(x: int) -> int:
        return x * 2

    cases = deal.cases(double2)
    case = next(iter(cases))
    with pytest.raises(deal.ExampleContractError):
        case()


@pytest.mark.skipif(hypothesis is None, reason='hypothesis is not installed')
def test_concat():
    from examples.concat import concat

    cases = deal.cases(concat)
    cases()



import ast

import pytest

from deal.linter._extractors import uses_result


try:
    import astroid
except ImportError:
    astroid = None


@pytest.mark.parametrize('given, expected', [
    # positive
    ('lambda a, result: 0', True),
    ('lambda a, result: 0', True),
    ('lambda a, *, result: 0', True),
    ('lambda *a, result: 0', True),
    ('lambda *a, result=None: 0', True),
    ('lambda _: _.result', True),
    ('lambda _: len(_.result) == _.length', True),

    # non-lambdas are allowed
    ('unknown', True),
    ('sum', True),

    # negative
    ('lambda _: 0', False),
    ('lambda _: _', False),
    ('lambda _: _.other', False),
    ('lambda a: 0', False),
    ('lambda a, res: 0', False),
    ('lambda a, *, res: 0', False),
])
def test_uses_result(given: str, expected: bool):
    if astroid is not None:
        tree = astroid.parse(given)
        print(tree.repr_tree())
        validator = tree.body[0].value
        assert uses_result(validator) is expected

    ast_tree = ast.parse(given)
    print(ast.dump(ast_tree))
    validator = ast_tree.body[0].value  # type: ignore[attr-defined]
    assert uses_result(validator) is expected



from io import StringIO
from pathlib import Path

import pytest

from deal._cli import main
from deal._cli._lint import LintCommand


try:
    import pygments
except ImportError:
    pygments = None

TEXT = """
import deal

@deal.post(lambda x: x > 0)
def f(x):
    return -1
"""


def test_get_errors(tmp_path: Path):
    (tmp_path / 'example.py').write_text(TEXT)
    errors = list(LintCommand.get_errors(paths=[str(tmp_path)]))
    assert len(errors) == 1
    assert errors[0]['code'] == 'DEL012'
    assert errors[0]['content'] == '    return -1'


@pytest.mark.skipif(pygments is None, reason='pygments is not installed')
def test_lint_command_colors(tmp_path: Path):
    (tmp_path / 'example.py').write_text(TEXT)
    stream = StringIO()
    count = main(['lint', str(tmp_path)], stream=stream)
    assert count == 1

    stream.seek(0)
    captured = stream.read()
    assert '\x1b[34mreturn\x1b[39;49;00m -\x1b[34m1\x1b[39;49;00m' in captured
    assert '\x1b[95m(-1)\x1b[0m' in captured
    assert '\x1b[95m^\x1b[0m' in captured


def test_lint_command_no_color(tmp_path: Path):
    (tmp_path / 'example.py').write_text(TEXT)
    stream = StringIO()
    count = main(['lint', '--nocolor', str(tmp_path)], stream=stream)
    assert count == 1

    stream.seek(0)
    captured = stream.read()
    exp = '6:11 DEL012 post contract error (-1) return -1 ^'
    assert captured.split()[1:] == exp.split()


def test_lint_command_two_files(tmp_path: Path):
    (tmp_path / 'example1.py').write_text(TEXT)
    (tmp_path / 'example2.py').write_text(TEXT)
    stream = StringIO()
    count = main(['lint', '--nocolor', str(tmp_path)], stream=stream)
    assert count == 2

    stream.seek(0)
    captured = stream.read()
    assert 'example1.py' in captured
    assert 'example2.py' in captured
    assert 'return -1' in captured
    assert '(-1)' in captured
    assert '^' in captured


def test_lint_command_two_errors(tmp_path: Path):
    (tmp_path / 'example.py').write_text('from deal import pre\n' + TEXT)
    stream = StringIO()
    count = main(['lint', '--nocolor', str(tmp_path)], stream=stream)
    assert count == 2

    stream.seek(0)
    captured = stream.read()
    assert 'return -1' in captured
    assert '(-1)' in captured
    assert '^' in captured


def test_lint_command_json(tmp_path: Path):
    (tmp_path / 'example.py').write_text(TEXT)
    stream = StringIO()
    count = main(['lint', '--json', str(tmp_path)], stream=stream)
    assert count == 1

    stream.seek(0)
    captured = stream.read()
    assert '"    return -1"' in captured
    assert '"-1"' in captured
    assert '^' not in captured



import deal


def inherit_method() -> None:
    class A:
        def f(self, x: int):
            raise NotImplementedError

    class B(A):
        @deal.inherit
        def f(self, x: int) -> float:
            raise NotImplementedError

    reveal_type(B().f)  # R: def (x: builtins.int) -> builtins.float


def inherit_class() -> None:
    class A:
        def f(self, x: int):
            raise NotImplementedError

    @deal.inherit
    class B(A):
        def f(self, x: int) -> float:
            raise NotImplementedError

    reveal_type(B().f)  # R: def (x: builtins.int) -> builtins.float



import ast
from textwrap import dedent

import pytest

from deal.linter._contract import Category
from deal.linter._extractors import get_contracts
from deal.linter._extractors.contracts import SUPPORTED_CONTRACTS, SUPPORTED_MARKERS


try:
    import astroid
except ImportError:
    astroid = None


def test_supported_contracts_match_categories():
    cats = {f'deal.{c.value}' for c in Category}
    assert cats == SUPPORTED_CONTRACTS | SUPPORTED_MARKERS


def get_cats(target) -> tuple:
    return tuple(contract.name for contract in get_contracts(target))


@pytest.mark.parametrize('text, expected', [
    ('@deal.post(lambda x: x>0)', ('post', )),
    ('@deal.pure', ('pure', )),
    ('@deal.pure()', ('pure', )),
    ('@deal.chain(deal.pure)', ('pure', )),
    ('@deal.chain(deal.pure, deal.post(lambda x: x>0))', ('pure', 'post')),
])
def test_decorators(text, expected):
    text += '\ndef f(x): pass'
    if astroid is not None:
        tree = astroid.parse(text)
        assert get_cats(tree.body[-1]) == expected

    ast_tree = ast.parse(text)
    print(ast.dump(ast_tree))
    assert get_cats(ast_tree.body[-1]) == expected


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_infer():
    text = """
        from io import StringIO
        import deal

        contracts = deal.chain(deal.pure, deal.post(lambda x: x>0))

        @contracts
        def f(x):
            return x

        with StringIO() as contracts2:
            @contracts2
            def f(x):
                return x
    """
    tree = astroid.parse(dedent(text))
    assert get_cats(tree.body[-2]) == ('pure', 'post')


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_infer_inherit_method():
    text = """
        import deal

        class B:
            @deal.has()
            def f(self):
                pass

        class A:
            def f(self):
                pass

        class C(A, B):
            @deal.pre(lambda: True)
            def f(self):
                pass

            @deal.raises(ZeroDivisionError)
            def f2(self):
                pass

        class D(Unknown, C):
            @deal.inherit
            @deal.post(lambda x: x>0)
            def f(self):
                return 2
    """
    tree = astroid.parse(dedent(text))
    cls = tree.body[-1]
    assert isinstance(cls, astroid.ClassDef)
    assert get_cats(cls.body[0]) == ('inherit', 'pre', 'has', 'post')


@pytest.mark.skipif(astroid is None, reason='astroid is not installed')
def test_inherit_no_parents():
    text = """
        import deal
        class A:
            @deal.inherit
            @deal.has()
            def f(self):
                pass
    """
    tree = astroid.parse(dedent(text))
    cls = tree.body[-1]
    assert isinstance(cls, astroid.ClassDef)
    assert get_cats(cls.body[0]) == ('inherit', 'has')


def test_inherit_function():
    text = dedent("""
        import deal
        @deal.inherit
        def f(x):
            pass
    """)
    if astroid is not None:
        tree = astroid.parse(text)
        assert get_cats(tree.body[-1]) == ('inherit',)

    ast_tree = ast.parse(text)
    assert get_cats(ast_tree.body[-1]) == ('inherit',)



import re

import deal


def contract(template: str, *args):
    rex = re.compile(r'\{\:([a-z])\}')
    types = {'s': str, 'd': float}
    matches = rex.findall(template)
    if len(matches) != len(args):
        return f'expected {len(matches)} argument(s) but {len(args)} found'
    for match, arg in zip(matches, args):
        t = types[match[0]]
        if not isinstance(arg, t):
            return f'expected {t.__name__}, {type(arg).__name__} given'
    return True


@deal.pre(contract)
def format(template: str, *args) -> str:
    return template.format(*args)


@deal.has('io')
def example():
    # good
    print(format('{:s}', 'hello'))

    # bad
    print(format('{:s}'))               # not enough args
    print(format('{:s}', 'a', 'b'))     # too many args
    print(format('{:d}', 'a'))          # bad type


if __name__ == '__main__':
    print(format('{:s} {:s}', 'hello', 'world'))
